# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

# pylint: disable=too-many-lines
'''This module contains classes that are supposed to act as containers of all
the information that is necessary to represent a test in a given format. For
instance, in the case of tables this includes the column contents, the headers,
etc.  It does **not** include any formatting information, such as column
widths, floating-point precision, colours, etc.  Decisions about the formatting
are handled by suitable formatting classes, such as :class:`~.Rst`.

.. _numpy indexing: https://numpy.org/doc/stable/user/basics.indexing.html
'''
import logging
import numpy as np


LOGGER = logging.getLogger(__name__)


class TableTemplate:
    '''A container class that encapsulates all the necessary information to
    represent a table.

    Examples of use of mainly show in context of concatentation of
    :class:`TableTemplate`, obtained with the :meth:`join` method.

    >>> import numpy as np
    >>> tit1 = TableTemplate(np.float_(1.5), np.float_(1.4),
    ...                      headers=['egg', 'spam'])
    >>> tit2 = TableTemplate(np.float_(1.2), np.float_(0.9),
    ...                      headers=['egg', 'spam'])
    >>> stab12 = join(tit1, tit2)
    >>> print(len(tit1.columns), len(tit2.columns))
    2 2
    >>> print(tit1.columns[0].size, tit2.columns[0].size)
    1 1
    >>> print(len(stab12.columns))
    2
    >>> print(stab12.columns[0].size)
    2
    >>> print(f"{stab12!r}")
    class: <class 'valjean.javert.templates.TableTemplate'>
    headers: ['egg', 'spam']
    egg: [1.5 1.2]
    spam: [1.4 0.9]
    highlights: [array([0., 0.]), array([0., 0.])]

    ``stab12`` contained both ``tit1`` and ``tit2`` as expected. Headers of the
    columns are the same, length of the columns is the sum of the two.

    >>> tit3 = TableTemplate(np.float_(0.8), np.float_(1.1),
    ...                      headers=['knight', 'parrot'])
    >>> stab13 = join(tit1, tit3)
    Traceback (most recent call last):
        ...
    ValueError: TableTemplates to add should have same headers

    An error is raised as the two :class:`TableTemplate` don't contain the same
    headers, so not the same kind of columns, thus they cannot be concatenated.

    It is also possible to join tables with same headers but different 'types'
    (scalars and arrays):

    >>> tit4 = TableTemplate(np.arange(4), np.arange(4)*0.5,
    ...                      headers=['egg', 'spam'])
    >>> print(len(tit4.columns), tit4.columns[0].size)
    2 4
    >>> stab14 = join(tit1, tit4)
    >>> print(len(stab14.columns), stab14.columns[0].size)
    2 5
    >>> stab14.columns[0].size == tit1.columns[0].size + tit4.columns[0].size
    True
    >>> print(f"{stab14!r}")
    class: <class 'valjean.javert.templates.TableTemplate'>
    headers: ['egg', 'spam']
    egg: [1.5 0.  1.  2.  3. ]
    spam: [1.4 0.  0.5 1.  1.5]
    highlights: [array([0., 0., 0., 0., 0.]), array([0., 0., 0., 0., 0.])]


    It is also possible to join arrays, a bigger array is obtained, without
    separation between the initial :class:`TableTemplate`:

    >>> tit5 = TableTemplate(np.arange(3)*0.1, np.arange(3)*0.05,
    ...                      headers=['egg', 'spam'])
    >>> stab45 = join(tit4, tit5)
    >>> print(len(stab45.columns), len(stab45.columns[0]))
    2 7
    >>> print(f"{stab45!r}")
    class: <class 'valjean.javert.templates.TableTemplate'>
    headers: ['egg', 'spam']
    egg: [0.  1.  2.  3.  0.  0.1 0.2]
    spam: [0.   0.5  1.   1.5  0.   0.05 0.1 ]
    highlights: [array([0., 0., 0., 0., 0., 0., 0.]), \
array([0., 0., 0., 0., 0., 0., 0.])]

    Any number of :class:`TableTemplate` can be joined (if fulfilling the
    requirements).

    >>> stab145 = join(tit1, tit4, tit5)
    >>> print(f"{stab145!r}")
    class: <class 'valjean.javert.templates.TableTemplate'>
    headers: ['egg', 'spam']
    egg: [1.5 0.  1.  2.  3.  0.  0.1 0.2]
    spam: [1.4  0.   0.5  1.   1.5  0.   0.05 0.1 ]
    highlights: [array([0., 0., 0., 0., 0., 0., 0., 0.]), \
array([0., 0., 0., 0., 0., 0., 0., 0.])]

    The :meth:`TableTemplate.join` method updates the left
    :class:`TableTemplate` as expected:

    >>> tit1.join(tit4, tit5)
    >>> print(f"{tit1!r}")
    class: <class 'valjean.javert.templates.TableTemplate'>
    headers: ['egg', 'spam']
    egg: [1.5 0.  1.  2.  3.  0.  0.1 0.2]
    spam: [1.4  0.   0.5  1.   1.5  0.   0.05 0.1 ]
    highlights: [array([0., 0., 0., 0., 0., 0., 0., 0.]), \
array([0., 0., 0., 0., 0., 0., 0., 0.])]
    '''

    def __init__(self, *columns, headers=None, units=None,
                 highlights=None):
        '''Construct a table from a set of columns. The columns must be
        :class:`numpy.ndarray` objects, and they must all contain the same
        number of elements (same array *size*).

        Column headers may be specified using the `headers` argument; in this
        case, the number of headers must be equal to the number of columns.

        Column units can also be specified using the `units` argument. Again,
        you must pass as many units as there are columns.

        Finally, it is possible to specify which table elements should be
        highlighted. This is done by passing a list of lists (or
        :class:`numpy.ndarray`) to the `highlights` argument. Each element of
        the `highlights` (outer) list represents a table column and therefore
        must have the same shape as all the other columns; also, the length of
        `highlights` must be equal to the number of columns. Elements of the
        inner lists (or :class:`numpy.ndarray`) must be booleans and indicate
        whether the corresponding table element must be highlighted.

        :param columns: a list of columns.
        :type columns: list(numpy.ndarray)
        :param list(str) headers: a list of headers.
        :param list(str) units: a list of measurement units.
        :param highlights: a list describing which table elements should be
            highlighted.
        :type highlights: list(list(bool)) or list(numpy.ndarray(bool))
        '''
        self.columns = columns
        n_columns = len(columns)
        self.headers = ['']*n_columns if headers is None else headers
        self.units = ['']*n_columns if units is None else units

        if not columns:
            raise ValueError('at least one column expected')

        if highlights is None:
            self.highlights = [np.full_like(self.columns[0], False)
                               for _ in self.columns]
        else:
            self.highlights = highlights

        if len(self.headers) != len(self.columns):
            err = (f'number of column headers ({len(self.headers)}) must '
                   f'match number of columns ({len(self.columns)})')
            raise ValueError(err)

        if len(self.units) != len(self.columns):
            err = (f'number of column units ({len(self.units)}) must match '
                   f'number of columns ({len(self.columns)})')
            raise ValueError(err)

        if not isinstance(self.highlights, list):
            raise TypeError('expected a list as the `highlights` argument, '
                            f'got a {type(self.highlights)} instead')
        if len(self.highlights) != len(self.columns):
            raise ValueError(f'expected a list of {len(self.columns)} '
                             'elements as the `highlights` argument, got '
                             f'{len(self.highlights)} elements instead')

        for i, col in enumerate(self.columns):
            if not isinstance(col, (np.ndarray, np.generic, list)):
                raise TypeError('table columns must be lists or Numpy arrays, '
                                f'not {type(col)}')

            col_size = (col.size if isinstance(col, (np.ndarray, np.generic))
                        else len(col))
            if i == 0:
                n_elems = col_size
            elif col_size != n_elems:
                err = ('columns must have the same number of elements; '
                       f'column 0 has size {n_elems}, but column {i} has size '
                       f'{col_size}')
                raise ValueError(err)

    def copy(self):
        '''Copy a :class:`TableTemplate` object.

        :rtype: TableTemplate

        .. note:: the highlignt function is not really copied, it has the same
            address as the self one. I don't know how to change that.
        '''
        return TableTemplate(*tuple(col.copy() for col in self.columns),
                             headers=self.headers.copy(),
                             units=self.units.copy(),
                             highlights=self.highlights.copy())

    def _binary_join(self, other):
        '''Join another :class:`TableTemplate` to the current one.

        This method **concatenates** :class:`TableTemplate` of the same number
        of columns and same headers. It returns an update of the left one.

        If the two :class:`TableTemplate` are not compatible an exception is
        raised.

        :param other: :class:`TableTemplate` to be added to the current one
        :raises TypeError: if the other parameter is not a
            :class:`TableTemplate`
        :raises ValueError: if the TableTemplates don't have the same headers.
        '''
        if not isinstance(other, TableTemplate):
            raise TypeError("Only a TableTemplate can be joined to another "
                            "TableTemplate")
        if self.headers != other.headers:
            raise ValueError("TableTemplates to add should have same headers")
        self.columns = tuple(np.hstack((self.columns[i], other.columns[i]))
                             for i in range(len(self.columns)))
        self.units = other.units if self.units is None else self.units
        self.highlights = list(np.hstack((
            [np.atleast_1d(h) for h in self.highlights],
            [np.atleast_1d(h) for h in other.highlights])))

    def join(self, *others):
        '''Join a given number a :class:`TableTemplate` to the current one.

        Only :class:`TableTemplate` with the same number of columns and same
        headers can be joined. The method returns the updated current one.

        :param others: list of TableTemplates to be joined to the current
            TableTemplate
        :type others: list(TableTemplate)
        '''
        for oti in others:
            self._binary_join(oti)

    def __repr__(self):
        '''Print TableTemplate details.'''
        intro = [f"class: {self.__class__}\nheaders: {self.headers}\n"]
        elts = []
        for header, col in zip(self.headers, self.columns):
            elts.append(f"{header}: {col}\n")
        hlight = [f'highlights: {self.highlights}']
        return ''.join(intro + elts + hlight)

    def __getitem__(self, index):
        '''Build a sliced :class:`TableTemplate` from the current
        :class:`TableTemplate`.

        Slicing is done like in the usual `NumPy` arrays, see:
        `numpy indexing`_ for more informations. No treatment like in
        :mod:`~valjean.eponine.dataset` is done.

        :param index: index, slice or tuple of slices
        :type index: int, slice, tuple(slice)
        :rtype: TableTemplate
        '''
        LOGGER.debug("In TableTemplate.__getitem__")
        if not isinstance(self.columns[0], np.ndarray):
            raise TypeError("Slicing is only possible when the values if the "
                            "TableTemplate is np.ndarray.")
        return TableTemplate(*tuple(col[index] for col in self.columns),
                             headers=self.headers.copy(),
                             units=self.units.copy(),
                             highlights=self.highlights.copy())

    def data(self):
        '''Yield bytes representing `self`. Two :class:`TableTemplate` objects
        containing equal data yield the same data.'''
        for col, head, unit, high in zip(self.columns, self.headers,
                                         self.units, self.highlights):
            yield np.require(col, requirements='C').data.cast('b')
            yield head.encode('utf-8')
            yield unit.encode('utf-8')
            yield high.data.cast('b')

    def __eq__(self, other):
        '''Test for equality of `self` and another :class:`TableTemplate`.'''
        return (len(self.columns) == len(other.columns)
                and all(np.array_equal(this, that)
                        for this, that in zip(self.columns, other.columns))
                and self.headers == other.headers
                and self.units == other.units
                and self.highlights == other.highlights)

    def __ne__(self, other):
        '''Test for inequality of `self` and another :class:`TableTemplate`.'''
        return not self == other


class CurveElements:
    '''Define the characteristics of a curve to plot.'''

    def __init__(self, values, bins, legend, *, index=0, errors=None):
        '''Construction of :class:`CurveElements`: curve details (values, bins,
        etc).

        Values and errors (if given) should be :obj:`numpy.ndarray` of same
        shape (they must have only non-trivial dimension).

        Bins are stored as a list of :obj:`numpy.ndarray`. This list should
        have the same length as the dimension of the values.

        The index is used to share the plotting style between curves that
        should. For example, if on a plot there are the reference and two
        curves representing different data, let's say 'egg' and 'spam', if we
        also want to draw the ratio of these data with the reference, the same
        style will be applied to 'egg vs reference' and 'egg' and to 'spam vs
        reference' and 'spam'. In that case to ensure the same style 'egg vs
        reference' and 'egg' should have the same index (same for the 'spam'
        case).

        :param numpy.ndarray values: array to be represented on the plot,
            **mandatory**
        :param list(numpy.ndarray) bins: bins to be used to represent the
            values
        :param str legend: string to be used in the legend to characterize the
            curve, **mandatory**
        :param int index: index of the curve (used for style for example)
        :param numpy.ndarray errors: errors associated to values (per default
            only on 1D plots and y-axis)
        '''
        self.values = values
        self.bins = bins
        self.legend = legend
        self.index = index
        self.errors = errors

        if not isinstance(self.values, np.ndarray):
            raise TypeError('Values should be np.ndarray.')

        if not len(bins) == self.values.ndim:
            raise ValueError('len(bins) should be equal to values dimension')

        if self.errors is not None:
            if not isinstance(self.errors, np.ndarray):
                raise TypeError('Errors should be np.ndarray or None')

    def copy(self):
        '''Copy a :class:`CurveElements` object.

        :rtype: CurveElements
        '''
        return CurveElements(values=self.values.copy(),
                             legend=self.legend,
                             bins=[b.copy() for b in self.bins],
                             index=self.index,
                             errors=(None if self.errors is None
                                     else self.errors.copy()))

    def __repr__(self):
        '''Printing of :class:`CurveElements`.'''
        elts = (f"  legend:  {self.legend}\n"
                f"  index:   {self.index}\n"
                f"  bins:    {self.bins}\n"
                f"  values:  {self.values}\n"
                f"  errors:  {self.errors}\n")
        return elts

    def __str__(self):
        '''Printing of :class:`CurveElements`.'''
        elts = (f"  legend:  {self.legend}\n"
                f"  index:   {self.index}\n"
                f"  bins:    {self.bins}\n")
        return elts

    def data(self):
        '''Generator yielding objects supporting the buffer protocol that (as a
        whole) represent a serialized version of `self`.'''
        yield np.require(self.values, requirements='C').data.cast('b')
        yield self.legend.encode('utf-8')
        for bins in self.bins:
            yield np.require(bins, requirements='C').data.cast('b')
        yield bytes((self.index,))
        if self.errors is not None:
            yield np.require(self.errors, requirements='C').data.cast('b')

    def __eq__(self, other):
        '''Test for equality of `self` and another :class:`CurveElements`.'''
        return (np.array_equal(self.values, other.values)
                and self.legend == other.legend
                and np.array_equal(self.bins, other.bins)
                and self.index == other.index
                and np.array_equal(self.errors, other.errors))

    def __ne__(self, other):
        '''Test for inequality of `self` and another :class:`CurveElements`.'''
        return not self == other


class SubPlotAttributes:
    # pylint: disable=too-many-instance-attributes
    '''Container to store sub-plots attributes:

    * axis limits
    * axis scale: linear (default) or logatithmic
    * additional horizontal or vertical lines

    Theses attributes are independent of the used backend (examples:
    **matplotlib**, **Root**, **gnuplot**, **D3**). The backend then gets the
    attributes and apply them with its own features.

    .. note:

        If needed new attributes can be added.
    '''
    def __init__(self, dim):
        '''Initialisation of PlotAttributes.

        The attributes of the instance are private.

        :param int dim: dimension of the data on the sub-plot (used to check
            consistency of limits)
        '''
        self.dim = dim
        self.logx = False
        self.logy = False
        self.logz = False
        self._limits = None
        self._lines = None

    def copy(self):
        '''Copy a :class:`SubPlotAttributes` object.

        :rtype: SubPlotAttributes
        '''
        spa = SubPlotAttributes(self.dim)
        spa.logx = self.logx
        spa.logy = self.logy
        spa.logz = self.logz
        spa.limits = self.limits.copy() if self.limits is not None else None
        spa.lines = self.lines.copy() if self.lines is not None else None
        return spa

    @property
    def limits(self):
        '''Return limits.'''
        return self._limits

    @limits.setter
    def limits(self, limits):
        '''Store limits for curves in the :class:`PlotTemplate`.

        :param list(tuple) limits: limits for each axis of the sub-plot

        Each tuple corresponds to an axis, 2 values are expected: min and max.
        '''
        if limits is not None:
            if len(limits) != self.dim:
                LOGGER.warning('Wrong number of limits given: expected %s, '
                               'got %s', self.dim, len(limits))
                return
        self._limits = limits

    @property
    def lines(self):
        '''Return lines to be plotted.'''
        return self._lines

    @lines.setter
    def lines(self, lines):
        '''Set lines to be added to the plots.

        :param list(list(dict)) lines: lines to be added to the plots
        '''
        if lines is None:
            self._lines = None
            return
        for line in lines:
            if len({'x', 'y'}.intersection(line)) != 1:
                LOGGER.warning('Only vertical or horizontal lines accepted')
                return
        self._lines = lines


class SubPlotElementsException(Exception):
    '''Error raised if the sub plot looks inconsistent.'''


class SubPlotElements:
    '''Container to store a given sub-plot.'''

    def __init__(self, *, curves, axnames=('', ''), ptype='1D'):
        '''Initialisation of :class:`SubPlotElements`.

        A subplot is defined as data (``curves``) sharing the same plotting
        properties:

        * axis names
        * type of the plot (e.g. ``'1D'``, ``'2D'``, ...), see the chosen
          backend to get the list of possibilities (example:
          :class:`~.mpl.MplPlot`)
        * axis scales
        * axis limits

        The last axis name corresponds to the quantity to be drawn, the first
        ones to the bins.

        Axis scales can be linear (default) or logarithmic (if set to
        ``True``).

        Vertical or horizontal lines can also be added.

        :param list(CurveElements) curves: list of curves to go on the sub-plot
        :param tuple axnames: name of the axes of the sub-plot
        :param str type: type of the sub-plot, default: ``'1D'``
        '''
        self.curves = curves
        self.axnames = list(axnames)
        self.ptype = ptype
        self._check_axes_consistency()
        self.attributes = SubPlotAttributes(len(self.axnames)-1)

    def _check_axes_consistency(self):
        '''Check axes consistency.

        Expected:

        * number of axis names = dimension of the curve + 1
        * if 1D: 2 axis names, if 2D: 3 axis names
        '''
        for crv in self.curves[1:]:
            if len(crv.bins) != len(self.curves[0].bins):
                raise SubPlotElementsException(
                    'Not the same number of axes in the curves')
        if len(self.axnames) != len(self.curves[0].bins)+1:
            raise SubPlotElementsException(
                'Inconsistent number of axis names and bins')
        if self.ptype == '1D' and len(self.axnames) != 2:
            raise SubPlotElementsException(
                f'Expecting a 1D plot but got {len(self.axnames)} axes')
        if self.ptype == '2D' and len(self.axnames) != 3:
            raise SubPlotElementsException(
                f'Expecting a 2D plot but got {len(self.axnames)} axes')
        if self.ptype == '2D' and len(self.curves) > 1:
            LOGGER.warning('Only one subplot expected by SubPlotElements in '
                           '2D case')

    def copy(self):
        '''Copy a :class:`SubPlotElements` object.

        :rtype: SubPlotElements
        '''
        celt = SubPlotElements(curves=[c.copy() for c in self.curves],
                               axnames=list(self.axnames),
                               ptype=self.ptype,)
        celt.attributes = self.attributes.copy()
        return celt

    def __repr__(self):
        '''Printing of :class:`SubPlotElements`'''
        elts = [f" axnames: {self.axnames}, plot type: {self.ptype}, "
                f"N curves: {len(self.curves)}\n"]
        for j, curve in enumerate(self.curves):
            elts.append(f" Curve {j}\n{curve!r}")
        return ''.join(elts)

    def __str__(self):
        '''Printing of :class:`SubPlotElements`'''
        elts = [f" axnames: {self.axnames}, plot type: {self.ptype}\n"]
        for j, curve in enumerate(self.curves):
            elts.append(f" Curve {j}\n{curve!s}")
        return ''.join(elts)

    def data(self):
        '''Generator yielding objects supporting the buffer protocol that (as a
        whole) represent a serialized version of `self`.'''
        for aname in self.axnames:
            yield aname.encode('utf-8')
        yield self.ptype.encode('utf-8')
        for curve in self.curves:
            yield from curve.data()

    def __eq__(self, other):
        '''Test of equality of `self` and another :class:`SubPlotElements`.'''
        return (all(s == o for s, o in zip(self.curves, other.curves))
                and self.axnames == other.axnames
                and self.ptype == other.ptype)

    def __ne__(self, other):
        '''Test for inequality of `self` and another :class:`SubPlotElements`.
        '''
        return not self == other


class PlotTemplate:
    '''A container for full test result to be represented as a plot. This
    includes all the datasets and the test result. This can also include
    p-values or other test parameters, depending on what is given to the
    :class:`PlotTemplate`.

    Examples mainly present the :meth:`join` method, used to concatentate
    :class:`PlotTemplate`.

    >>> bins1, d11, d12 = np.arange(4), np.arange(4), np.arange(4)*10
    >>> d13 = d11 + d12
    >>> bins2, d2 = np.arange(5), np.arange(5)*0.5
    >>> pit1 = PlotTemplate(subplots=[SubPlotElements(
    ...     curves=[CurveElements(d11, [bins1], 'd11', index=0)],
    ...     axnames=['egg', 'brandy'])])
    >>> pit2 = PlotTemplate(subplots=[SubPlotElements(
    ...     curves=[CurveElements(d12, bins=[bins1], legend='d12', index=1)],
    ...     axnames=['egg', 'beer'])])
    >>> pit3 = PlotTemplate(subplots=[SubPlotElements(
    ...     curves=[CurveElements(d13, legend='d13', bins=[bins1], index=2)],
    ...     axnames=['egg', 'wine'])])
    >>> splt123 = join(pit1, pit2, pit3)
    >>> print(f"{splt123!r}")
    class:   <class 'valjean.javert.templates.PlotTemplate'>
    N subplots: 3
    Subplot 0
     axnames: ['egg', 'brandy'], plot type: 1D, N curves: 1
     Curve 0
      legend:  d11
      index:   0
      bins:    [array([0, 1, 2, 3])]
      values:  [0 1 2 3]
      errors:  None
    Subplot 1
     axnames: ['egg', 'beer'], plot type: 1D, N curves: 1
     Curve 0
      legend:  d12
      index:   1
      bins:    [array([0, 1, 2, 3])]
      values:  [ 0 10 20 30]
      errors:  None
    Subplot 2
     axnames: ['egg', 'wine'], plot type: 1D, N curves: 1
     Curve 0
      legend:  d13
      index:   2
      bins:    [array([0, 1, 2, 3])]
      values:  [ 0 11 22 33]
      errors:  None
    <BLANKLINE>

    As expected a new :class:`PlotTemplate` is obtained, containing three
    subplots, each one containing one curve.

    Like in the TableTemplate case, the :meth:`PlotTemplate.join` method
    updates the left :class:`PlotTemplate` as expected:

    >>> pit1.join(pit2, pit3)
    >>> pit1 == splt123
    True

    A new curve with the same axes will also create a new suplot:

    >>> d14 = d11*2
    >>> pit4 = PlotTemplate(subplots=[SubPlotElements(
    ...     curves=[CurveElements(d14, legend='d14', index=3, bins=[bins1])],
    ...     axnames=['egg', 'beer'])])
    >>> split24 = join(pit2, pit4)
    >>> print(split24)
    class:   <class 'valjean.javert.templates.PlotTemplate'>
    Subplot 0
     axnames: ['egg', 'beer'], plot type: 1D
     Curve 0
      legend:  d12
      index:   1
      bins:    [array([0, 1, 2, 3])]
    Subplot 1
     axnames: ['egg', 'beer'], plot type: 1D
     Curve 0
      legend:  d14
      index:   3
      bins:    [array([0, 1, 2, 3])]
    <BLANKLINE>


    To get it in the same subplot it has to be done at creation.

    >>> pit24 = PlotTemplate(subplots=[SubPlotElements(
    ...     curves=[CurveElements(d12, legend='d12', index=1, bins=[bins1]),
    ...             CurveElements(d14, legend='d14', index=3, bins=[bins1])],
    ...     axnames=['egg', 'beer'])])
    >>> print(pit24)
    class:   <class 'valjean.javert.templates.PlotTemplate'>
    Subplot 0
     axnames: ['egg', 'beer'], plot type: 1D
     Curve 0
      legend:  d12
      index:   1
      bins:    [array([0, 1, 2, 3])]
     Curve 1
      legend:  d14
      index:   3
      bins:    [array([0, 1, 2, 3])]
    <BLANKLINE>
    >>> split24 == pit24
    False


    N-dimensional plot templates can be built, but the plotting engine may not
    be able to convert multi-dimensional templates into plots.

    The same behavior is expected for multi-dimensions plots: ``join`` will had
    a new subplot. To be noted: only one curve can be plotted on a subplot in
    multi-dimensional case, so any additional curve will throw a warning. Plot
    representation may not be as expected.

    >>> d31 = np.arange(bins1.size*bins2.size).reshape(bins1.size, bins2.size)
    >>> d32 = np.arange(bins1.size*bins2.size).reshape(
    ...     bins1.size, bins2.size)*0.01
    >>> pit7 = PlotTemplate(subplots=[SubPlotElements(
    ...     curves=[CurveElements(d31, bins=[bins1, bins2],
    ...                           legend='d31', index=0)],
    ...     axnames=['egg', 'spam', 'bacon'], ptype='2D')])
    >>> pit8 = PlotTemplate(subplots=[SubPlotElements(
    ...     curves=[CurveElements(d32, bins=[bins1, bins2],
    ...                           legend='d32', index=1)],
    ...     axnames=['egg', 'spam', 'lobster'], ptype='2D')])
    >>> splt78 = join(pit7, pit8)
    >>> print(f"{splt78!s}")
    class:   <class 'valjean.javert.templates.PlotTemplate'>
    Subplot 0
     axnames: ['egg', 'spam', 'bacon'], plot type: 2D
     Curve 0
      legend:  d31
      index:   0
      bins:    [array([0, 1, 2, 3]), array([0, 1, 2, 3, 4])]
    Subplot 1
     axnames: ['egg', 'spam', 'lobster'], plot type: 2D
     Curve 0
      legend:  d32
      index:   1
      bins:    [array([0, 1, 2, 3]), array([0, 1, 2, 3, 4])]
    <BLANKLINE>

    It is also possible to mix 1D and 2D plots:

    >>> splt27 = join(pit2, pit7)
    >>> print(f"{splt27!s}")
    class:   <class 'valjean.javert.templates.PlotTemplate'>
    Subplot 0
     axnames: ['egg', 'beer'], plot type: 1D
     Curve 0
      legend:  d12
      index:   1
      bins:    [array([0, 1, 2, 3])]
    Subplot 1
     axnames: ['egg', 'spam', 'bacon'], plot type: 2D
     Curve 0
      legend:  d31
      index:   0
      bins:    [array([0, 1, 2, 3]), array([0, 1, 2, 3, 4])]
    <BLANKLINE>

    '''
    def __init__(self, *, subplots, small_subplots=True, suppress_xaxes=False,
                 suppress_legends=False, backend_kw=None):
        '''Construction of the PlotTemplate from a list of
        :class:`SubPlotElements`.

        :param list(SubPlotElements) subplots: list of sub-plots
        :param bool small_subplots: draw additional subplots in smaller size
            than the first one, default = ``True``
        :param bool suppress_xaxes: suppress label and ticks labels of the
            x-axis of subplots except the last one, default = ``False``
        :param bool suppress_legends: suppress legend on all subplots except
            the first one, default = ``False``
        :param dict backend_kw: dictionary with backend-specific options
        '''
        self.subplots = subplots
        self.nb_plots = len(self.subplots)
        self.small_subplots = small_subplots
        self.suppress_xaxes = suppress_xaxes
        self.suppress_legends = suppress_legends
        self.backend_kw = backend_kw if backend_kw else {}

        if not isinstance(self.subplots, list):
            raise TypeError("The 'subplots' argument must a list of "
                            "SubPlotElements.")

        if not all(isinstance(spl, SubPlotElements) for spl in self.subplots):
            raise TypeError("The 'subplots' argument must a list of "
                            "SubPlotElements.")

    def copy(self):
        '''Copy a :class:`PlotTemplate` object.

        :rtype: PlotTemplate
        '''
        ccplt = PlotTemplate(subplots=[pelt.copy() for pelt in self.subplots],
                             small_subplots=self.small_subplots,
                             suppress_xaxes=self.suppress_xaxes,
                             suppress_legends=self.suppress_legends,
                             backend_kw=self.backend_kw)
        return ccplt

    def _binary_join(self, other):
        '''Join another :class:`PlotTemplate` to the current one.

        This method **concatenates** the current :class:`PlotTemplate` to the
        ``other`` one, i.e., if they share the axes (bins and name), the curves
        list is extended.

        The curves are not added in the mathematical sense. If this is what you
        really want to do, please add the datasets themselves before and
        probably redo the test.

        :param other: PlotTemplate to be added to the current one
        :type other: PlotTemplate
        :raises TypeError: if the other parameter is not a
            :class:`PlotTemplate`
        :raises ValueError: if axes are not the same (``axname`s` and ``bins``)
        '''
        if not isinstance(other, PlotTemplate):
            raise TypeError("Only a PlotTemplate can be joined to another "
                            "PlotTemplate")
        self.subplots.extend(other.subplots)
        self.nb_plots = len(self.subplots)
        if other.small_subplots != self.small_subplots:
            LOGGER.warning('Not same value for small_subplots in self and '
                           'other, keeping self one (%s)', self.small_subplots)
        if other.suppress_xaxes != self.suppress_xaxes:
            LOGGER.warning('Not same value for suppress_xaxes in self and '
                           'other, keeping self one (%s)', self.suppress_xaxes)
        if other.suppress_legends != self.suppress_legends:
            LOGGER.warning('Not same value for suppress_legends in self and '
                           'other, keeping self one (%s)',
                           self.suppress_legends)
        if other.backend_kw != self.backend_kw:
            LOGGER.warning('Not same values for backend_kw in self and other,'
                           ' keeping self one (%s)', self.backend_kw)

    def join(self, *others):
        '''Join a given number a :class:`PlotTemplate` to the current one.

        Only :class:`PlotTemplate` with the same number of columns and same
        headers can be joined. The method returns the updated current one.

        :param others: list of PlotTemplates to be join with the current
            PlotTemplate
        :type others: list(PlotTemplate)
        '''
        for oti in others:
            self._binary_join(oti)

    def __repr__(self):
        '''Printing of :class:`PlotTemplate`.'''
        elts = [f"class:   {self.__class__}\nN subplots: {self.nb_plots}\n"]
        for i, splt in enumerate(self.subplots):
            elts.append(f"Subplot {i}\n{splt!r}")
        return ''.join(elts)

    def __str__(self):
        '''Printing of :class:`PlotTemplate`.'''
        elts = [f"class:   {self.__class__}\n"]
        for i, splt in enumerate(self.subplots):
            elts.append(f"Subplot {i}\n{splt!s}")
        return ''.join(elts)

    def data(self):
        '''Yield bytes representing `self`. Two :class:`TableTemplate` objects
        containing equal data yield the same data.'''
        for splt in self.subplots:
            yield from splt.data()

    def __eq__(self, other):
        '''Test for equality of `self` and another :class:`PlotTemplate`.'''
        return self.subplots == other.subplots

    def __ne__(self, other):
        '''Test for inequality of `self` and another :class:`PlotTemplate`.'''
        return not self == other

    def curves_index(self):
        '''Return a sorted list of unique index of the curves.'''
        lind = []
        for splt in self.subplots:
            for crv in splt.curves:
                if crv.index not in lind:
                    lind.append(crv.index)
        return lind


class TextTemplate:
    '''A container class that encapsulates text for the report.

    The user has to write the text as a string. ReST markdown can be used as
    compilation is expected to be done by sphinx.

    .. note::

        Titles might not be well represented in a ReSt formatted TextTemplate:
        no knowledge of the current level of title, nor the associated symbol.
        Lists or enumerations might be more suitable.

    As in the other templates, examples will focus on the concatenation (join)
    of different TextTemplate.

    >>> ttplt1 = TextTemplate('Spam egg bacon')
    >>> print(f"{ttplt1!r}")
    <class 'valjean.javert.templates.TextTemplate'>\
(text='Spam egg bacon')
    >>> ttplt2 = TextTemplate('**Spam** egg bacon')
    >>> print(f"{ttplt2!r}")
    <class 'valjean.javert.templates.TextTemplate'>\
(text='**Spam** egg bacon')
    >>> ttplt3 = TextTemplate(r".. role:: hl\\n\\nsausage :hl:`tomato`")
    >>> ttplt1.join(ttplt3)
    >>> print(f"{ttplt1!r}")
    <class 'valjean.javert.templates.TextTemplate'>\
(text='Spam egg bacon.. role:: hl\\\\n\\\\nsausage :hl:`tomato`')

    The sphinx compilation will fail there, as the are no empty line between
    the first and the second string. If you know some text will follow, think
    about the ``\\n``.

    >>> ttplt1 = TextTemplate('Spam egg bacon\\n\\n')
    >>> ttplt3 = TextTemplate('.. role:: hl\\n\\nsausage :hl:`tomato`\\n\\n')
    >>> ttplt1.join(ttplt3)
    >>> print(f"{ttplt1!r}")
    <class 'valjean.javert.templates.TextTemplate'>\
(text='Spam egg bacon\\n\\n.. role:: hl\\n\\nsausage :hl:`tomato`\\n\\n')

    Test of the external function :meth:`join`:

    >>> ttplt4 = join(ttplt3, ttplt3)
    >>> print(f"{ttplt4!r}")
    <class 'valjean.javert.templates.TextTemplate'>\
(text='.. role:: hl\\n\\nsausage :hl:`tomato`\\n\\n\
.. role:: hl\\n\\nsausage :hl:`tomato`\\n\\n')

    The copy doesn't affect the original:

    >>> ttplt5 = ttplt2.copy()
    >>> ttplt5.text += ' sausage'
    >>> print(f"{ttplt2!r}")
    <class 'valjean.javert.templates.TextTemplate'>\
(text='**Spam** egg bacon')
    >>> print(f"{ttplt5!r}")
    <class 'valjean.javert.templates.TextTemplate'>\
(text='**Spam** egg bacon sausage')
    '''

    def __init__(self, text):
        '''Construct the text to be sent to the report.

        :param str text: text to be written in the report
        '''
        self.text = text

    def __repr__(self):
        '''Print :class:`TextTemplate` details.'''
        return f'{self.__class__}(text={self.text!r})'

    def _binary_join(self, other):
        if not isinstance(other, TextTemplate):
            raise TypeError("Only a TextTemplate can be joined to another "
                            f"TextTemplate; got a {type(other)}")
        self.text += other.text

    def copy(self):
        '''Copy a :class:`TextTemplate` object.

        :rtype: TextTemplate
        '''
        return TextTemplate(text=self.text)

    def join(self, *others):
        '''Join a given number of :class:`TextTemplate` to the current one.'''
        for oti in others:
            self._binary_join(oti)

    def data(self):
        '''Yield bytes representing `self`. Two :class:`TableTemplate` objects
        containing equal data yield the same data.'''
        yield self.text.encode('utf-8')

    def __eq__(self, other):
        '''Test for equality of `self` and another :class:`TextTemplate`.'''
        return self.text == other.text

    def __ne__(self, other):
        '''Test for inequality of `self` and another :class:`TextTemplate`.
        '''
        return not self == other


def join(*templates):
    '''Join a "list" of templates of same kind, :class:`TableTemplate` or
    :class:`PlotTemplate` using the related ``join`` methods.

    It returns a new templates (:class:`TableTemplate` or
    :class:`PlotTemplate`).

    :param templates: list of templates
    :type templates: list(TableTemplate) or list(PlotTemplate)
    :rtype: TableTemplate, PlotTemplate, TextTemplate


    See :class:`TableTemplate` and :class:`PlotTemplate` for examples of use.
    Only few error cases will be shown here:

    >>> bins1, data11, data12 = np.arange(4), np.arange(4), np.arange(4)*10
    >>> bins2, data2 = np.arange(5), np.arange(5)*0.5
    >>> tablit = TableTemplate(bins1, data11, headers=['egg', 'spam'])
    >>> plotit = PlotTemplate(subplots=[SubPlotElements(
    ...     curves=[CurveElements(data11, bins=[bins1], legend='d11')],
    ...     axnames=['egg', 'spam'])])
    >>> tit = join(tablit, plotit)
    Traceback (most recent call last):
        ...
    TypeError: Only a TableTemplate can be joined to another TableTemplate
    >>> tit = join(plotit, tablit)
    Traceback (most recent call last):
        ...
    TypeError: Only a PlotTemplate can be joined to another PlotTemplate
    '''
    copy = templates[0].copy()
    copy.join(*templates[1:])
    return copy



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''General-purpose fingerprinting.

This module provides a single function, called :func:`fingerprint`, that
computes a SHA-256 hash for any object implementing a `self.data()` generator.
The `self.data()` method is expected to yield bytestrings representing the
internal state of the object. The hash computed by :func:`fingerprint`
represents the state of the object in a compact way.
'''

from hashlib import sha256


def fingerprint(obj):
    '''Compute the fingerprint of `obj`.

    Example:

    >>> class MyObject:
    ...     def __init__(self, x):
    ...         self.x = x
    ...     def data(self):
    ...         yield str(self.x).encode('utf-8')
    >>> obj = MyObject(42)
    >>> fingerprint(obj)
    '73475cb40a568e8da8a045ced110137e159f890ac4da883b6b17dc651b3a8049'
    >>> obj = MyObject('some_string')
    >>> fingerprint(obj)
    '539a374ff43dce2e894fd4061aa545e6f7f5972d40ee9a1676901fb92125ffee'

    :returns: a hash representing `obj`
    :rtype: str
    '''
    hasher = sha256()
    for data in obj.data():
        hasher.update(data)
    return hasher.hexdigest()



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

r'''
In our case we'll use the modified χ²-test widely used in physics for comparing
histograms. This test is also available in the **ROOT**
`software <https://root.cern.ch/>`_. For further explanations a good desciption
is provided by the documentation of ROOT's ``Chi2Test()`` method on
`TH1 page <https://root.cern.ch/doc/master/classTH1.html>`_ or in this
`article <https://arxiv.org/abs/physics/0605123>`_ by N. Gagunashvili.

The used formula is then:

.. math::

    \chi^2_{obs} = \sum_{i=1}^p\frac{(x_{1i} - x_{2i})^2}
    {\sigma_{1i}^2 + \sigma_{2i}^2}

where datasets (or distributions) 1 and 2 are compared, :math:`x_i` being the
values and :math:`\sigma_i` being the error in each bin.

In our case only the 'weighted' case is implemented.


Statistical interpretation
``````````````````````````

The null hypothesis for the χ²-test is:

* equal means in each bin
* gaussian means in each bin (so normal distribution in each bin)
* bins are independent

The interpretation of the test is based on the χ² value and on the p-value.
Usually it is expected that the χ²/ndf should be close to 1 (ndf = number of
degrees of freedom), see `PDG, statistics chapter
<https://pdg.lbl.gov/2018/reviews/rpp2018-rev-statistics.pdf>`_. The p-value
can also be calculated from the probability density function of the χ²
distribution, which is given by:

.. math::
    pdf(x,k) = \frac{1}{2^{k/2}\Gamma(k/2)}x^{k/2-1}e^{-x/2}

for :math:`x > 0`, :math:`k` the number of degrees of freedom and
:math:`\Gamma` is Euler's Gamma function [#GammaFunc]_.

The cumulative distribution function, :math:`cdf`, used to calculated the
p-value is then:

.. math::

    cdf(x,k) = \frac{\gamma(k/2,x/2)}{\Gamma(k/2)}
             = \int_{-\infty}^{x}pdf(t,k)dt

where :math:`\gamma` is the lower incomplete gamma function
[#partialGammaFunc]_.

It is equivalent to calculate the probability :math:`P(x < \chi^2_{obs})` which
is equal to :math:`1 - cdf(x,k)`, so the integral from :math:`x` to
:math:`+\infty`. This probability is the p-value.

If the p-value is lower than the define level (1%, 5% for example), the null
hypothesis is rejected.

Possible biases can come from wrong estimation of uncertainties
(underestimated for example or only statistical while a systematic one
should be taken into account).


Code implementation
```````````````````

**SciPy** is used to calculate the p-value. The distribution
:obj:`scipy.stats.chi2` is imported as `schi2`` in order to name conflicts. The
p-value is obtained with the ``sf`` method. It is directly compared with the
significance level required by the user.

.. warning::

    If some of your bins are not filled, if they are initialized to zero or if
    their error is zero, the χ² calculation will be wrong and probably give
    :math:`\infty`.

    One option is proposed to remove these bins from the calculation,
    ``ignore_empty``. The use of this option implies a reduction of the number
    of degrees of freedom by the number of removed bins (ndf = # bins). It has
    to be considered with caution, this does not make necessarly sense,
    depending on the user case.

    Another option is to apply some rebinning in order to have no empty bins
    (what usually means re-running to get consistent uncertainties).


Examples of the test
````````````````````

Let's consider a spectrum of 5 bins with their error and apply the χ²-test.


>>> from valjean.eponine.dataset import Dataset
>>> from valjean.gavroche.stat_tests.chi2 import TestChi2
>>> import numpy as np
>>> ds1 = Dataset(np.array([5.2, 5.3, 5.25, 5.4, 5.5]),
...               np.array([0.2, 0.25, 0.1, 0.2, 0.3]))
>>> ds2 = Dataset(np.array([5.1, 5.6, 5.2, 5.3, 5.2]),
...               np.array([0.1, 0.3, 0.05, 0.4, 0.3]))
>>> tchi2 = TestChi2(ds1, ds2, alpha=0.05, name="comp",
...                  description="Comparison using Chi2 test")
>>> tchi2_res = tchi2.evaluate()
>>> print(f'{tchi2_res.chi2_per_ndf[0]:.7f}')
0.3080328
>>> bool(tchi2_res)
True
>>> print(f'{tchi2_res.pvalue[0]:.7f}')
0.9083889


Test with empty bins
^^^^^^^^^^^^^^^^^^^^

One "empty bin" at the same position in the 2 compared datasets prevents the
test to work. An empty bin is defined as a bin with zero error. In Monte
Carlo codes a zero error can happen for example when only one batch has been
run, no variance can be calculated in that case (``n-1`` usually used with
``n`` the number of batches).

>>> ds3 = Dataset(np.array([5.2, 5.3, 5.25, 5.4, 5.5]),
...               np.array([0.2, 0.25, 0., 0.2, 0.3]))
>>> ds4 = Dataset(np.array([5.1, 5.6, 5.2, 5.3, 5.2]),
...               np.array([0.1, 0.3, 0., 0.4, 0.3]))
>>> tchi2 = TestChi2(ds3, ds4, alpha=0.05, name="comp",
...                  description="Comparison using Chi2 test")
>>> tchi2_res = tchi2.evaluate()
>>> tchi2_res.chi2
[inf]
>>> bool(tchi2_res)
False
>>> print(np.array2string(tchi2_res.test.nonzero_bins[0]))
[ True  True  True  True  True]
>>> tchi2_res.test.ndf[0]
5
>>> np.count_nonzero(ds4.error)
4

If the values are also at zero, we get a ``nan`` instead of an ``inf``.

To get a χ² evaluation based on the non-empty bins, use the ``ignore_empty``
argument of the :class:`TestChi2`. It recalculates the number of degrees of
freedom removing the empty bins and only use the non-zero ones in the χ²
calculation.

>>> tchi2 = TestChi2(ds3, ds4, alpha=0.05, ignore_empty=True,
...                  name="comp", description="Comparison using Chi2 test")
>>> tchi2_res = tchi2.evaluate()
>>> print(f'{tchi2_res.chi2[0]:.7f}')
1.3401639
>>> tchi2_res.test.ndf[0]
4
>>> tchi2_res.test.dsref.size
5
>>> print(np.array2string(tchi2_res.test.nonzero_bins[0]))
[ True  True False  True  True]
>>> print(f'{tchi2_res.chi2_per_ndf[0]:.7f}')
0.3350410
>>> bool(tchi2_res)
True


Test with multiple dimensions datasets
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

>>> ds5 = Dataset(np.array([[5.2, 5.3, 5.25], [5.4, 5.5, 5.2]]),
...               np.array([[0.2, 0.25, 0.1], [0.2, 0.3, 0.1]]))
>>> ds6 = Dataset(np.array([[5.1, 5.6, 5.2], [5.3, 5.2, 5.3]]),
...               np.array([[0.1, 0.3, 0.05], [0.4, 0.3, 0.2]]))
>>> ds5.shape
(2, 3)
>>> ds5.size
6
>>> tchi2 = TestChi2(ds5, ds6, alpha=0.05, name="comp",
...                  description="Comparison using Chi2 test")
>>> tchi2_res = tchi2.evaluate()
>>> print(f'{tchi2_res.chi2_per_ndf[0]:.7f}')
0.2900273
>>> bool(tchi2_res)
True
>>> print(f'{tchi2_res.pvalue[0]:.7f}')
0.9419786

And when we have at least one empty bin, using the ``ignore_empty`` argument:

>>> ds7 = Dataset(np.array([[5.2, 5.3, 5.25], [5.4, 5.5, 5.2]]),
...               np.array([[0.2, 0.25, 0.], [0.2, 0.3, 0.1]]))
>>> ds8 = Dataset(np.array([[5.1, 5.6, 5.2], [5.3, 5.2, 5.4]]),
...               np.array([[0.1, 0.3, 0.], [0.4, 0.3, 0.]]))
>>> ds7.shape
(2, 3)
>>> ds7.size
6
>>> tchi2 = TestChi2(ds7, ds8, alpha=0.05, ignore_empty=True,
...                  name="comp", description="Comparison using Chi2 test")
>>> tchi2_res = tchi2.evaluate()
>>> print(f'{tchi2_res.chi2[0]:.7f}')
5.3401639
>>> tchi2_res.test.ndf[0]
5
>>> tchi2_res.test.dsref.size
6
>>> print(np.array2string(tchi2_res.test.nonzero_bins[0]))
[[ True  True False]
 [ True  True  True]]
>>> print(f'{tchi2_res.chi2_per_ndf[0]:.7f}')
1.0680328
>>> bool(tchi2_res)
True

.. note::

    A RuntimeWarning is emitted if zero bins are used during calculation.


.. rubric:: Footnotes

.. [#GammaFunc] The Γ function is given by

                .. math::

                    \Gamma(s) = \int_0^\infty t^{s-1}e^{-t}dt = (s-1)!

                It also has an incomplete form (not starting at 0):

                .. math::

                    \Gamma(s, x) = \int_x^\infty t^{s-1}e^{-t}dt

.. [#partialGammaFunc] The γ function is given by:

                .. math::

                    \gamma(s, x) = \int_0^x t^{s-1}e^{-t}dt

'''
import numpy as np
from scipy.stats import chi2 as schi2
from ..test import TestDataset, TestResult


class TestResultChi2(TestResult):
    '''Result from the χ²-test.'''

    def __init__(self, test, chi2, pvalue):
        '''Construct a TestResultChi2 object.

        Members are lists which length corresponds to the number of datasets
        compared to the reference dataset.

        :param test: the χ² test object
        :type test: :class:`~.TestChi2`
        :param chi2: the value of the χ² statistic
        :type chi2: :class:`list` (:obj:`numpy.generic` (:obj:`float`))
        :param pvalue: the test p-value
        :type pvalue: :class:`list` (:obj:`numpy.generic` (:obj:`float`))
        '''
        super().__init__(test)
        self.chi2 = chi2
        self.pvalue = pvalue

    def oracles(self):
        '''Final test for the list of compared datasets.

        :returns: :obj:`numpy.ndarray`
        '''
        return np.greater(self.pvalue, self.test.alpha)

    def __bool__(self):
        return all(self.oracles())

    @property
    def chi2_per_ndf(self):
        '''Calculate the χ² per number of degrees of freedom.

        :returns: χ²/ndf
        :rtype: float

        No parenthesis needed: this is a property.
        '''
        return [c2 / df for c2, df in zip(self.chi2, self.test.ndf)]


class TestChi2(TestDataset):
    '''Test class for χ², inheritate from :class:`~valjean.gavroche.test.Test`.
    '''

    def __init__(self, dsref, *datasets, name, description='', labels=None,
                 alpha=0.01, ignore_empty=False):
        # pylint: disable=too-many-arguments
        '''Initialisation of :class:`TestChi2`

        :param str name: local name of the test
        :param str description: specific description of the test
        :param dict labels: labels to be used for test classification in
                            reports (for example category, input file name,
                            type of result, ...)
        :param dsref: reference dataset
        :type dsref: :class:`~valjean.eponine.dataset.Dataset`
        :param datasets: list of datasets  to be compared to reference dataset
        :type datasets: :class:`list` (:class:`~.dataset.Dataset`)
        :param float alpha: probability to accept of not the test (p-value is
                            expected greater), i.e. significance level
        :param bool ignore_empty: if the array contains zero bins, the test can
                                  done not consider them if this option is
                                  True; otherwise it will probably fail as the
                                  array to sum will contain infinite terms.
                                  Default is ``False``.
        '''
        super().__init__(dsref, *datasets,
                         name=name, description=description, labels=labels)
        self.alpha = alpha
        self.ignore_empty = ignore_empty
        #: nonzero bins identification by True, False if zero
        #: (:class:`list` (:obj:`numpy.ndarray` (:obj:`bool`)))
        self.nonzero_bins = self._nonzero_bins()
        #: number of degrees of freedom (:class:`list` (:obj:`int`))
        self.ndf = [np.count_nonzero(nzb) for nzb in self.nonzero_bins]

    def _nonzero_bins(self):
        '''Identify nonzero bins.

        A nonzero bin is considered as a bin where at least one of the datasets
        has a nonzero error. Consequently if both bins have zero error this is
        a zero bin. This identification is stored in ``nonzero_bins``: ``True``
        for nonzero, else ``False``. This :obj:`numpy.ndarray` is used in the
        χ²-test calculation.

        :returns: :obj:`numpy.ndarray` (:obj:`bool`)
        '''
        if self.ignore_empty:
            return [np.logical_or(self.dsref.error > 0, _ds.error > 0)
                    for _ds in self.datasets]
        return [np.full_like(self.dsref.value, True, dtype=bool)
                for _ in range(len(self.datasets))]

    @staticmethod
    def pvalue(chi2, ndf):
        '''Calculation of the p-value of the test.

        :param chi2: observed χ²
        :type chi2: :obj:`numpy.generic` (:obj:`float`)
        :param ndf: number of degrees of freedom
        :type ndf: int or :obj:`numpy.generic` (:obj:`int`)
        :returns: :obj:`numpy.generic` (:obj:`float`)
        '''
        return schi2.sf(chi2, ndf)

    @staticmethod
    def chi2_test(ds1, ds2, nonzero_bins=None):
        r'''Compute the χ² value for the given datasets.

        :param ds1: first dataset
        :type ds1: :class:`~valjean.eponine.dataset.Dataset`
        :param ds2: second dataset
        :type ds2: :class:`~valjean.eponine.dataset.Dataset`
        :param nonzero_bins: optional argument, booleans array of the same size
                             as ``ds1`` and ``ds2`` to identify zero bins and
                             possibly avoid them (see below)
        :type nonzero_bins: :obj:`numpy.ndarray` (:obj:`bool`) or ``None``
                            (default)
        :returns: χ² value
        :rtype: :obj:`numpy.generic` (:obj:`float`)

        It has to be noted that if in a bin the sum of the variances of the
        test and the reference dataset is zero (so both datasets bin has a zero
        error) the bin cannot be taken into account in the calculation of the
        sum as the ratio will be infinite. This is why they can be removed via
        the ``nonzero_bins`` argument at that step. This has another
        consequence: the number of degrees of freedom is consequently reduced
        by the number of bins removed.
        '''
        diff = ds1 - ds2
        if nonzero_bins is None:
            return np.sum(((diff.value / diff.error)**2))
        return np.sum(((diff.value / diff.error)**2)[nonzero_bins])

    def evaluate(self):
        '''Evaluate χ² test method.

        :rtype: :class:`~.TestResultChi2`
        '''
        chi2 = []
        for _ds, _nzb in zip(self.datasets, self.nonzero_bins):
            chi2.append(self.chi2_test(self.dsref, _ds, _nzb))
        pvalue = [self.pvalue(_chi2, _ndf)
                  for _chi2, _ndf in zip(chi2, self.ndf)]
        return TestResultChi2(self, chi2, pvalue)

    def data(self):
        '''Generator yielding objects supporting the buffer protocol that (as a
        whole) represent a serialized version of `self`.'''
        yield from super().data()
        yield self.__class__.__name__.encode('utf-8')
        yield float(self.alpha).hex().encode('utf-8')
        n_bytes = (self.ignore_empty.bit_length() - 1) // 8 + 1
        yield self.ignore_empty.to_bytes(n_bytes, 'little')



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

r'''Module to deal with ROOT outputs from Tripoli-4 in depletion mode.

.. _the ROOT website: https://root.cern/


This module is an interface with Tripoli-4 depletion results, stored as
``ROOT`` files, see the Tripoli-4 user guide and `the ROOT website`_ for more
details.

The available results match the usual Tripoli-4 results from depletion:

* k\ :sub:`eff` as ``kcoll``, ``ktrack``, ``kstep``;
* β\_:sub:`eff` from prompt and Nauchi, called ``beff_prompt`` and
  ``beff_nauchi``;
* renormalisation factor, called ``renorm``
* total power
* power
* local burnup
* flux: fast neutron flux (``fast_flux``), thermal neutron flux
  (``therm_flux``) and total flux (``flux``)
* mass
* concentration
* activity
* reaction rates: for fast neutrons (``fast_reaction_rate``), for thermal
  neutrons (``thermal_reaction_rate``) and for one group (``reaction_rate``).
  The thermal group only gives non-zero results for fission rate.

Some of these results can be accessed from valjean using the
:class:`DepletionReader` class from this module. In some cases, it is necessary
to provide keyword for the composition name (``componame``), the isotope name
(``isotope``) or the reaction name (``reaction``).

Results can be obtained at a given depletion step (so a scalar value) or for
all available steps, giving in both cases a
:class:`~valjean.eponine.dataset.Dataset`. The access method of the array ends
with ``_time`` if the required quantity is given in function of *time* and
``_burnup`` if it is in function of *burnup*.

A typical example of use of this module is the following:

.. code-block:: python

    from valjean.eponine.tripoli4.depletion import DepletionReader
    depr = DepletionReader.from_evolution_steps(
        'evolution_1.root', 'evolution_2.root',
        root_build='/path/where/to/build/root/libraries')
    # get kstep at step 1
    kstep = depr.kstep(1)
    # get kstep as a function of burnup
    akstep = depr.kstep_burnup()
    # get total power as a function of time
    atotpow = depr.total_power_time()
    # get U238 concentration in composition COMPO1 at step 5
    conc_u238 = depr.concentration(step=5, componame='COMPO1', isotope='U238')
    # get U238 fission reaction rate in composition COMP1 as a function of time
    reac_u238 = depr.reaction_rate_time(componame='COMP1', isotope='U238',
                                        reaction='REAMT18')

Some helper methods provide the list of compositions, the list of isotopes in a
given composition or the reaction rates associated to a given isotope in a
given composition.
'''

# pylint: disable=no-member

import logging
import re
from collections import OrderedDict
from pathlib import Path
import pkg_resources as pkg
import numpy as np
from ..dataset import Dataset


LOGGER = logging.getLogger(__name__)


def title_to_snake_case(word):
    '''Convert `word` from title case to snake case.

    >>> title_to_snake_case('ThisIsATitle')
    'this_is_a_title'
    '''
    word = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', word)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', word).lower()


def generic_docstrings(res_type, *list_params):
    '''Define generic docstrings for the access functions of depleted results.
    '''
    docstrings = []
    if res_type == 'value':
        docstrings.extend([
            'Return the value of {thing} from the MeanBurnupResults object.\n',
            ':param int step: the index of the calculation step'])
    else:
        if res_type == 'burnup':
            docstrings.extend([
                'Return the {thing} from the MeanBurnupResults object as a '
                'function of burnup.\n'])
        elif res_type == 'time':
            docstrings.extend([
                'Return the {thing} from the MeanBurnupResults object as a '
                'function of time.\n'])
        else:
            raise DepletionReaderException('Not foreseen result type: '
                                           f'{res_type}')
        docstrings.append(
            '''.. note::

            The unit of the x-axis is not currently preserved as units are not
            taken into account in :class:`~valjean.eponine.dataset.Dataset`.

            ''')

    docstrings_params = {
        'componame': ':param str componame: composition name for {thing}',
        'isotope': ':param str isotope: isotope which {thing} is required',
        'reaction': ':param str reaction: reaction name'}

    docstrings.extend([docstrings_params[lpar] for lpar in list_params])

    if res_type == 'value':
        docstrings.extend([
            ':returns: {thing} and error for the given step',
            ':rtype: Dataset'])
    elif res_type == 'burnup':
        docstrings.extend([
            ":returns: {thing} and error as a function of burnup steps",
            ":rtype: Dataset"])
    else:
        docstrings.extend([
            ":returns: {thing} and error as a function of time steps",
            ":rtype: Dataset"])

    return '\n'.join(docstrings)


def add_accessors(dict_res):
    '''Automatic construction of result accessors.'''

    def decorator(cls):
        def method_factory(name, *args):
            '''Produce methods for the decorated class.

            This function takes as an argument the name of a quantity to
            retrieve from the ``MeanBurnupResults`` object. It returns three
            methods (burnup, time and value) that can be added as
            accessors to `cls`.'''
            kwargs = {arg: None for arg in args}

            def burnup_method(self, **kwargs):
                mbr_method_name = 'Get' + name + 'Histogram'
                method = getattr(self.mbr, mbr_method_name)
                # burnup -> index=0
                hist = method(**kwargs, index=0)
                return Dataset(
                    value=np.array(hist.values.data()),
                    error=np.array(hist.errors.data()),
                    bins=OrderedDict([(hist.xname,
                                       np.array(hist.bins.data()))]),
                    what=hist.yname)
            burnup_method.__doc__ = generic_docstrings(
                'burnup', *args).format(thing=name)

            def time_method(self, **kwargs):
                mbr_method_name = 'Get' + name + 'Histogram'
                method = getattr(self.mbr, mbr_method_name)
                # time -> index=1
                hist = method(**kwargs, index=1)
                return Dataset(
                    value=np.array(hist.values.data()),
                    error=np.array(hist.errors.data()),
                    bins=OrderedDict([(hist.xname,
                                       np.array(hist.bins.data()))]),
                    what=hist.yname)
            time_method.__doc__ = generic_docstrings(
                'time', *args).format(thing=name)

            def value_method(self, step, **kwargs):
                mbr_method_name = 'Get' + name
                method = getattr(self.mbr, mbr_method_name)
                value = method(step, *list(kwargs.values()), 0)
                error = method(step, *list(kwargs.values()), 1)
                return Dataset(value=value, error=error)
            value_method.__doc__ = generic_docstrings(
                'value', *list(kwargs.keys())).format(thing=name)

            return burnup_method, time_method, value_method

        for name, args in dict_res.items():
            sc_name = title_to_snake_case(name)
            bu_method, time_method, value_method = method_factory(name, *args)
            setattr(cls, sc_name + '_burnup', bu_method)
            setattr(cls, sc_name + '_time', time_method)
            setattr(cls, sc_name, value_method)
        return cls
    return decorator


@add_accessors({
    'Kcoll': [],
    'Ktrack': [],
    'Kstep': [],
    'BeffPrompt': [],
    'BeffNauchi': [],
    'Renorm': [],
    'TotalPower': [],
    'Power': ['componame'],
    'LocalBurnup': ['componame'],
    'FastFlux': ['componame'],
    'ThermFlux': ['componame'],
    'Mass': ['componame', 'isotope'],
    'Concentration': ['componame', 'isotope'],
    'Activity': ['componame', 'isotope'],
    'ReactionRate': ['componame', 'isotope', 'reaction'],
    'FastReactionRate': ['componame', 'isotope', 'reaction'],
    'ThermalReactionRate': ['componame', 'isotope', 'reaction']})
class DepletionReader:
    '''Class to use depletion results from Tripoli-4'''

    def __init__(self, mbr):
        '''Initialisation of DepletionReader.

        :param mbr: MeanBurnupResult
        '''
        self.mbr = mbr

    @staticmethod
    def init_postscripts(root_build=""):
        '''Initialize postscripts from ROOT macros.

        ROOT macros are in the `resources/depletion` folder. They are compiled
        in the `__t4depletion__` folder to used afterwards.
        '''
        try:
            import ROOT
        except ImportError as ierr:
            LOGGER.error('ROOT needs to be added to PYTHONPATH')
            raise ImportError('ROOT missing') from ierr
        if all(hasattr(ROOT, cls) for cls in ['DepletedComposition',
                                              'BurnupResults',
                                              'MeanBurnupResults']):
            LOGGER.info('skipping compilation of the ROOT depletion scripts '
                        'because they are already available')
            return ROOT
        LOGGER.info('attempting compilation of the ROOT depletion scripts...')
        ps_fold = 'valjean.eponine.tripoli4.resources.depletion'
        assert pkg.resource_exists(ps_fold, 'DepletedComposition.C')
        assert pkg.resource_exists(ps_fold, 'BurnupResults.C')
        assert pkg.resource_exists(ps_fold, 'MeanBurnupResults.C')
        fname = pkg.resource_filename(ps_fold, 'DepletedComposition.C')
        path_fname = Path(fname)
        ROOT.gROOT.SetMacroPath(str(path_fname.parent))
        if not root_build:
            LOGGER.warning('T4 depletion postscripts ROOT libraries will be '
                           'compiled in valjean folder, permissions might not '
                           'be granted.')
            root_build = path_fname.parent / "__t4depletion__"
        ROOT.gSystem.MakeDirectory(str(root_build))
        ROOT.gSystem.SetBuildDir(str(root_build), True)
        ROOT.gROOT.LoadMacro('DepletedComposition.C+')
        ROOT.gROOT.LoadMacro('BurnupResults.C+')
        ROOT.gROOT.LoadMacro('MeanBurnupResults.C+')
        LOGGER.info('... ROOT depletion scripts have been compiled and loaded')
        return ROOT

    @classmethod
    def from_evolution_steps(cls, *fnames, root_build=""):
        '''Initialisation from `evolution.root` files (one per simulation).

        :param str fnames: ROOT files
        '''
        root = cls.init_postscripts(root_build)
        cls.fnames = fnames
        lmbr = root.MeanBurnupResults()
        for fname in fnames:
            lmbr.AddSimulationAndProcess(fname)
        return cls(mbr=lmbr)

    @classmethod
    def from_mbr(cls, fname, mbr_name, root_build=""):
        '''Initialisation from ROOT file containing a MeanBurnupResults class.

        :param str fname: ROOT file name
        :param str mbr_name: name of the MeanBurnupResults object in the file
        '''
        root = cls.init_postscripts(root_build)
        cls.fname = fname
        tfname = root.TFile(fname)
        lmbr = tfname.Get(mbr_name)
        tfname.Close()
        return cls(mbr=lmbr)

    def save_mbr(self, name):
        '''Save the MeanBurnupResults in a ROOT file.

        :param str name: name of the output ROOT file name
        '''
        self.mbr.SaveAs(name)

    def nb_simu(self):
        '''Return the number of independent simulations.

        :rtype: int
        '''
        return self.mbr.GetNbSimu()

    def nb_compositions(self):
        '''Return the number of compositions.

        :rtype; int
        '''
        return self.mbr.GetNbCompos()

    def nb_steps(self):
        '''Return the number of steps.

        :rtype: int
        '''
        return self.mbr.GetSteps()

    def burnup(self, step):
        r'''Return burnup from MeanBurnupResults object.

        :param int step: calculation step to use to get the value
        :returns: burnup value and error
        :rtype: Dataset
        '''
        return Dataset(value=self.mbr.GetBurnup(step, 0),
                       error=self.mbr.GetBurnup(step, 1))

    def burnup_array(self):
        r'''Return burnup array from MeanBurnupResults object.

        :returns: burnup value and error by step
        :rtype: Dataset
        '''
        val, err = [], []
        steps = list(range(1, self.mbr.GetSteps()+1))
        for i in steps:
            val.append(self.mbr.GetBurnup(i, 0))
            err.append(self.mbr.GetBurnup(i, 1))
        return Dataset(
            value=np.array(val), error=np.array(err),
            bins=OrderedDict([('step', np.array(steps))]))

    def time(self, step):
        r'''Return time from MeanBurnupResults object.

        :param int step: calculation step to use to get the value
        :returns: time value and error
        :rtype: Dataset
        '''
        return Dataset(value=self.mbr.GetTime(step, 0),
                       error=self.mbr.GetTime(step, 1))

    def time_array(self):
        r'''Return time array from MeanBurnupResults object.

        :returns: time value and error by step
        :rtype: Dataset
        '''
        val, err = [], []
        steps = list(range(1, self.mbr.GetSteps()+1))
        for i in steps:
            val.append(self.mbr.GetTime(i, 0))
            err.append(self.mbr.GetTime(i, 1))
        return Dataset(
            value=np.array(val), error=np.array(err),
            bins=OrderedDict([('step', np.array(steps))]))

    def composition_names(self):
        '''Return the list of composition names.

        :rtype: list(str)
        '''
        lcomp = [self.mbr.GetDepletedCompositionName(i)
                 for i in range(1, self.nb_compositions()+1)]
        return lcomp

    def isotope_names(self, step, componame):
        '''Return the list of isotopes in the given composition at the given
        step.

        :param int step: calculation step to use to get the value
        :param str componame: composition name in which list of isotopes is
            required
        :returns: list of available isotopes
        :rtype: list(str)
        '''
        depli = list(self.mbr.GetDepletedIsotopeNames(step, componame))
        return [str(iso) for iso in depli]

    def reaction_names(self, step, componame, isotope):
        '''Get reaction names in composition for the given isotope.

        :param int step: calculation step to use to get the value
        :param str componame: composition name where reaction is required
        :param str isotope: isotope name for which reaction is required
        :returns: list of available reactions
        :rtype: list(str)
        '''
        srn = self.mbr.GetDepletedReactionNames(step, componame, isotope)
        return [str(i) for i in srn]

    def isotope_reaction_names(self, step, componame):
        '''Get dictionary of reactions per isotope for the given step and
        composition.

        :param int step: calculation step to use to get the value
        :param str componame: composition name where reactions per isotope are
            required
        :returns: reactions by isotopes
        :rtype: dict(str, list(str))
        '''
        lir = self.mbr.GetDepletedIsotopeReactionNames(step, componame)
        return {i: [str(r) for r in setr] for i, setr in lir}

    def dump_global_results(self, step):
        '''Print results for a given step.

        This includes compositions.
        '''
        self.mbr.DumpGlobalResults(step)


class DepletionReaderException(Exception):
    '''An error that may be raised by the :class:`DepletionReader` class.'''



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

# pylint: disable=too-many-lines

'''This module provides the classes to convert test results to plots using
:mod:`matplotlib.pyplot`.

.. _legend documentation:
    https://matplotlib.org/api/_as_gen/matplotlib.axes.Axes.legend.html
.. _matplotlib colors:
    https://matplotlib.org/tutorials/colors/colors.html
.. _matplotlib markers:
    https://matplotlib.org/api/markers_api.html
.. _matplotlib styles:
    https://matplotlib.org/gallery/style_sheets/style_sheets_reference.html

:class:`MplPlot` objects take as input :class:`~.templates.PlotTemplate`
containing curves (:class:`~.templates.CurveElements`) classified by sub-plots
(:class:`~.templates.SubPlotElements`).

The format, or rendering, of the plot can be set using the rcParams but also
some predefined parameters on which the class cycle like colors, markers shape
and filling.

By default the first color is black and is used only once: it is excluded from
the cycle on colors. It is typically reserved for the reference but can be
reused if the first index is used for another curve.

Plots can be obtained with the following for example:

.. plot::
    :include-source:

    >>> import numpy as np
    >>> from valjean.javert.templates import (PlotTemplate, CurveElements,
    ...                                       SubPlotElements)
    >>> bins = (np.array(np.arange(10)),)
    >>> lcurves = [CurveElements(values=bins[0]*0.5*(icurve+1), bins=bins,
    ...                          legend=str(icurve+1), index=icurve)
    ...            for icurve in range(20)]
    >>> pltit = PlotTemplate(subplots=[SubPlotElements(
    ...     curves=lcurves, axnames=('the x-axis', ''), ptype='1D')])
    >>> from valjean.javert.mpl import MplPlot
    >>> mplplt = MplPlot(pltit)
    >>> fig, _ = mplplt.draw()

Additional subplots can be drawn if required. The style of the curves is fixed
by the index (see :class:`.CurveElements`).

.. _plot 3 panels:
.. plot::
    :include-source:

    >>> import numpy as np
    >>> from valjean.javert.templates import (PlotTemplate, CurveElements,
    ...                                       SubPlotElements)
    >>> bins = (np.array(np.arange(10)),)
    >>> lcurves1 = []
    >>> for icurve in range(3):
    ...     lcurves1.append(CurveElements(
    ...         values=bins[0][1:]*0.5*(icurve+1) + icurve*(-1)**(icurve),
    ...         bins=bins, legend=str(icurve), index=icurve))
    >>> sbpe1 = SubPlotElements(curves=lcurves1, axnames=('the x-axis', ''))
    >>> lcurves2 = []
    >>> for icurve in range(1, 3):
    ...     lcurves2.append(CurveElements(
    ...         values=lcurves1[icurve].values/lcurves1[0].values, bins=bins,
    ...         legend=str(icurve+1), index=icurve))
    >>> sbpe2 = SubPlotElements(curves=lcurves2,
    ...                         axnames=('the x-axis', 'C/ref'))
    >>> lcurves3 = []
    >>> for icurve in range(1, 3):
    ...     lcurves3.append(CurveElements(
    ...         values=((lcurves1[icurve].values-lcurves1[0].values)
    ...                 /lcurves1[0].values),
    ...         bins=bins, legend=str(icurve+1), index=icurve))
    >>> sbpe3 = SubPlotElements(curves=lcurves3,
    ...                         axnames=('the x-axis', '(C-ref)/ref'))
    >>> pltit = PlotTemplate(subplots=[sbpe1, sbpe2, sbpe3])
    >>> from valjean.javert.mpl import MplPlot
    >>> mplplt = MplPlot(pltit)
    >>> fig, _ = mplplt.draw()

These examples also show the default style of the plots.


Style setup
-----------

Some style parameters are available in the object :class:`MplStyle`: general
style, colors of markers and lines (expected to be the same for the same
curve), shapes and fills of markers. Legend keyword arguments can also be
given.


General style
`````````````

It is possible to change the general style of plots using a predefined one
or to use different markers. The predefined styles can be seen in `matplotlib
styles`_ or be obtained thanks to

>>> import matplotlib.pyplot as plt  # doctest: +SKIP
>>> print(plt.style.available)  # doctest: +SKIP

For example, we can have:

.. plot::
    :include-source:

    >>> import numpy as np
    >>> from valjean.javert.templates import (PlotTemplate, CurveElements,
    ...                                       SubPlotElements)
    >>> bins = (np.array(np.arange(10)),)
    >>> lcurves = []
    >>> for icurve in range(20):
    ...     lcurves.append(CurveElements(values=bins[0]*0.5*(icurve+1),
    ...                                  bins=bins, legend=str(icurve+1),
    ...                                  index=icurve))
    >>> sbpe = SubPlotElements(curves=lcurves, axnames=['the x-axis', ''])
    >>> pltit = PlotTemplate(subplots=[sbpe])
    >>> from valjean.javert.mpl import MplPlot, MplStyle
    >>> mplplt = MplPlot(pltit, style=MplStyle(style='Solarize_Light2'))
    >>> fig, _ = mplplt.draw()

Colors and markers
``````````````````

Colors and markers can also be changed directly:

.. plot::
    :include-source:

    >>> import numpy as np
    >>> from valjean.javert.templates import (PlotTemplate, CurveElements,
    ...                                       SubPlotElements)
    >>> bins = [np.array(np.arange(10))]
    >>> lcurves = []
    >>> for icurve in range(20):
    ...     lcurves.append(CurveElements(values=bins[0]*0.5*(icurve+1),
    ...                                  bins=bins, legend=str(icurve+1),
    ...                                  index=icurve))
    >>> sbpe = SubPlotElements(curves=lcurves, axnames=['the x-axis', ''])
    >>> pltit = PlotTemplate(subplots=[sbpe])
    >>> from valjean.javert.mpl import MplPlot, MplStyle
    >>> style = MplStyle(colors=['b', 'g', 'r', 'y', 'm'],
    ...                  mshape=['X', '+', 'D', '1', 'p', 'v', 'o'],
    ...                  mfill=['top', 'full', 'right', 'none', 'bottom',
    ...                         'left', 'none'])
    >>> mplplt = MplPlot(pltit , style=style)
    >>> fig, _ = mplplt.draw()


Legends
```````

By default the legend is represented on all panels at the location
:mod:`matplotlib.pyplot` determines like in `plot 3 panels`_.

If you would prefer to get only one legend for all panels, the
``suppress_legends`` argument in :class:`~.templates.PlotTemplate` should be
used. In that case, only fine for 1D plots, the legend will be placed on the
largest panel by default.

In the style any keyword argument accepted by matplotlib can be given to modify
for example the legend position. This is can be found in the `legend
documentation`_.

The next example show the `plot 3 panels`_ with only one legend which position
and shape have been modified.

.. plot::
    :include-source:

    >>> import numpy as np
    >>> from valjean.javert.templates import (PlotTemplate, CurveElements,
    ...                                       SubPlotElements)
    >>> bins = [np.array(np.arange(10))]
    >>> lcurves1 = []
    >>> for icurve in range(3):
    ...     lcurves1.append(CurveElements(
    ...         values=bins[0][1:]*0.5*(icurve+1) + icurve*(-1)**(icurve),
    ...         bins=bins, legend=str(icurve), index=icurve))
    >>> sbpe1 = SubPlotElements(curves=lcurves1, axnames=['the x-axis', ''])
    >>> lcurves2 = []
    >>> for icurve in range(1, 3):
    ...     lcurves2.append(CurveElements(
    ...         values=lcurves1[icurve].values/lcurves1[0].values,
    ...         bins=bins, legend=str(icurve+1)+' vs 0', index=icurve))
    >>> sbpe2 = SubPlotElements(curves=lcurves2,
    ...                         axnames=['the x-axis', 'C/ref'])
    >>> lcurves3 = []
    >>> for icurve in range(1, 3):
    ...     lcurves3.append(CurveElements(
    ...         values=((lcurves1[icurve].values-lcurves1[0].values)
    ...                 /lcurves1[0].values),
    ...         bins=bins, legend=str(icurve+1)+' vs 0', index=icurve))
    >>> sbpe3 = SubPlotElements(curves=lcurves3,
    ...                         axnames=['the x-axis', '(C-ref)/ref'])
    >>> pltit = PlotTemplate(subplots=[sbpe1, sbpe2, sbpe3],
    ...                      suppress_legends=True)
    >>> from valjean.javert.mpl import MplPlot, MplStyle
    >>> style = MplStyle(legends={'loc': 3, 'bbox_to_anchor': (0., 1., 1, 1),
    ...                           'mode': 'expand'})
    >>> mplplt = MplPlot(pltit, style=style)
    >>> fig, _ = mplplt.draw()


2D plots
--------

2D plots are also done via the class :class:`MplPlot`. The plot type ``ptype``
in :class:`~.templates.PlotTemplate` should be ``'2D'``. The principle is the
same as for 1D plots. Three axes are expected. Each curve has its own plot, no
superposition is done, so one subplot is expected to contain only one curve.
Each subplot can then have its own properties.

The colorbar axis label is set using the third axis name.

There is no real legend, so ``legend`` is used as title of each plot.

The index is currently not used.

.. plot::
    :include-source:

    >>> from collections import OrderedDict
    >>> import numpy as np
    >>> from valjean.javert.templates import (PlotTemplate, CurveElements,
    ...                                       SubPlotElements)
    >>> bins = [np.arange(6), np.arange(17, step=2)]
    >>> axnames = ['x', 'y']
    >>> incv = np.arange(1, 41).reshape(5, 8)
    >>> decv = np.arange(1, 41)[::-1].reshape(5, 8)
    >>> lsplts = []
    >>> lsplts.append(SubPlotElements(
    ...     curves=[CurveElements(
    ...         values=incv, bins=bins, legend='increase', index=0)],
    ...     axnames=['x', 'y', 'spam'], ptype='2D'))
    >>> lsplts.append(SubPlotElements(
    ...     curves=[CurveElements(
    ...         values=decv, bins=bins, legend='decrease', index=0)],
    ...     axnames=['x', 'y', 'spam'], ptype='2D'))
    >>> lsplts.append(SubPlotElements(
    ...     curves=[CurveElements(
    ...         values=incv/decv, bins=bins, legend='', index=1)],
    ...     axnames=['x', 'y', 'ratio'], ptype='2D'))
    >>> pltnd = PlotTemplate(subplots=lsplts, small_subplots=False)
    >>> from valjean.javert import mpl
    >>> mplplt = mpl.MplPlot(pltnd)
    >>> fig, _ = mplplt.draw()


.. note::

    Per default, additional subplots are small ones, it is probably better in
    2D case to set the parameter ``small_subplots`` to ``False`` in the
    :class:`~.templates.PlotTemplate`.

.. warning::

    Requesting more than one curve on a subplot will emit a warning but give
    unexpected results (typically only one of the 2D plot will be shown).

.. plot::
    :include-source:

    >>> from collections import OrderedDict
    >>> import numpy as np
    >>> from valjean.javert.templates import (PlotTemplate, CurveElements,
    ...                                       SubPlotElements)
    >>> bins = [np.arange(6), np.arange(17, step=2)]
    >>> axnames = ['x', 'y']
    >>> incv = np.arange(1, 41).reshape(5, 8)
    >>> decv = np.arange(1, 41)[::-1].reshape(5, 8)
    >>> lsplts = []
    >>> lsplts.append(SubPlotElements(
    ...     curves=[CurveElements(values=incv, bins=bins, legend='increase'),
    ...             CurveElements(values=decv, bins=bins, legend='decrease')],
    ...     axnames=['x', 'y', 'spam'], ptype='2D'))
    >>> lsplts.append(SubPlotElements(
    ...     curves=[CurveElements(
    ...         values=incv/decv, bins=bins, legend='', index=1)],
    ...     axnames=['x', 'y', 'ratio'], ptype='2D'))
    >>> pltnd = PlotTemplate(subplots=lsplts)
    >>> from valjean.javert import mpl
    >>> mplplt = mpl.MplPlot(pltnd)
    >>> fig, _ = mplplt.draw()


Customization
-------------

Some customizations can be done for each subplot with the attributes parameter
of :class:`~.templates.SubPlotElements`: limits to adapt axes ranges,
logarithmic scale or lines.

Using the previous 1D example:

.. plot::
    :include-source:

    >>> import numpy as np
    >>> from valjean.javert.templates import (PlotTemplate, CurveElements,
    ...                                       SubPlotElements)
    >>> bins = [np.array(np.arange(10))]
    >>> lcurves1 = [CurveElements(
    ...     values=bins[0][1:]*0.5*(icurve+1) + icurve*(-1)**(icurve),
    ...     bins=bins, legend=str(icurve), index=icurve)
    ...             for icurve in range(3)]
    >>> sbpe1 = SubPlotElements(curves=lcurves1, axnames=['the x-axis', ''])
    >>> lcurves2 = [CurveElements(
    ...     values=lcurves1[icurve].values/lcurves1[0].values, bins=bins,
    ...     legend=str(icurve+1)+' vs 0', index=icurve)
    ...             for icurve in range(1, 3)]
    >>> sbpe2 = SubPlotElements(curves=lcurves2,
    ...                         axnames=['the x-axis', 'C/ref'])
    >>> lcurves3 = [CurveElements(
    ...     values=((lcurves1[icurve].values-lcurves1[0].values)
    ...             /lcurves1[0].values),
    ...     bins=bins, legend=str(icurve+1)+' vs 0', index=icurve)
    ...             for icurve in range(1, 3)]
    >>> sbpe3 = SubPlotElements(curves=lcurves3,
    ...                         axnames=['the x-axis', '(C-ref)/ref'])
    >>> sbpe1.attributes.logx = True
    >>> sbpe2.attributes.limits = [(2, 7)]
    >>> sbpe3.attributes.logy = True
    >>> pltit = PlotTemplate(subplots=[sbpe1, sbpe2, sbpe3],
    ...                      small_subplots=False)
    >>> from valjean.javert.mpl import MplPlot, MplStyle
    >>> mplplt = MplPlot(pltit)
    >>> fig, _ = mplplt.draw()


Customizations specific to the backend, here :mod:`matplotlib.pyplot`, are
passed thanks to the ``backend_kw`` argument of
:class:`~.templates.PlotTemplate`. The keywords that are currently supported by
the :mod:`matplotlib.pyplot` backend and used in :mod:`valjean` are:

``nrows`` and ``ncols``
    These keys should be associated to integers. They determine the number of
    subplots, in case the :class:`~.templates.PlotTemplate` object contains
    several :class:`~.templates.SubPlotElements`. If ``ncols`` is given,
    ``nrows`` also has to be given.

``figsize``
    This key should be associated to tuple of floats (width and height in
    inches) to specify the figure size.

``subplot_kw``
    This key should be associated to a dictionary of keywords that will be
    passed as-is to :func:`matplotlib.pyplot.subplots`.

``gridspec_kw``
    This key should be associated to a dictionary to be passed to the grid
    constructor, see :func:`matplotlib.pyplot.subplots` and
    :class:`matplotlib.gridspec.GridSpec`.

All possible keyword arguments that can be passed to
:func:`matplotlib.pyplot.subplots` are normally supported.

Example is given from the previous 1D one:

.. plot::
    :include-source:

    >>> import numpy as np
    >>> from valjean.javert.templates import (PlotTemplate, CurveElements,
    ...                                       SubPlotElements)
    >>> bins = [np.array(np.arange(10))]
    >>> lcurves1 = [CurveElements(
    ...     values=bins[0][1:]*0.5*(icurve+1) + icurve*(-1)**(icurve),
    ...     bins=bins, legend=str(icurve), index=icurve)
    ...             for icurve in range(3)]
    >>> sbpe1 = SubPlotElements(curves=lcurves1, axnames=['the x-axis', ''])
    >>> lcurves2 = [CurveElements(
    ...     values=lcurves1[icurve].values/lcurves1[0].values, bins=bins,
    ...     legend=str(icurve+1)+' vs 0', index=icurve)
    ...             for icurve in range(1, 3)]
    >>> sbpe2 = SubPlotElements(curves=lcurves2,
    ...                         axnames=['the x-axis', 'C/ref'])
    >>> lcurves3 = [CurveElements(
    ...     values=((lcurves1[icurve].values-lcurves1[0].values)
    ...             /lcurves1[0].values),
    ...     bins=bins, legend=str(icurve+1)+' vs 0', index=icurve)
    ...             for icurve in range(1, 3)]
    >>> sbpe3 = SubPlotElements(curves=lcurves3,
    ...                         axnames=['the x-axis', '(C-ref)/ref'])
    >>> pltit = PlotTemplate(subplots=[sbpe1, sbpe2, sbpe3],
    ...                      small_subplots=False,
    ...                      backend_kw={'ncols': 3, 'nrows': 1,
    ...                                  'subplot_kw': {'aspect': 'equal'}})
    >>> from valjean.javert.mpl import MplPlot, MplStyle
    >>> mplplt = MplPlot(pltit)
    >>> fig, _ = mplplt.draw()


Customization also works on 2D plots. In addition the colorscale and colormap
can be put in logarithmic scale.

.. plot::
    :include-source:

    >>> from collections import OrderedDict
    >>> import numpy as np
    >>> from valjean.javert.templates import (PlotTemplate, CurveElements,
    ...                                       SubPlotElements)
    >>> bins = [np.arange(6), np.arange(17, step=2)]
    >>> axnames = ['x', 'y']
    >>> incv = np.arange(1, 41).reshape(5, 8)
    >>> decv = np.arange(1, 41)[::-1].reshape(5, 8)
    >>> sbp1 = SubPlotElements(
    ...     curves=[CurveElements(values=incv, bins=bins, legend='increase')],
    ...     axnames=['x', 'y', 'spam'], ptype='2D')
    >>> sbp2 = SubPlotElements(
    ...     curves=[CurveElements(values=decv, bins=bins, legend='decrease')],
    ...     axnames=['x', 'y', 'spam'], ptype='2D')
    >>> sbp3 = SubPlotElements(
    ...     curves=[CurveElements(values=incv/decv, bins=bins, legend='i/d')],
    ...     axnames=['x', 'y', 'ratio'], ptype='2D')
    >>> sbp3.attributes.logz = True
    >>> pltnd = PlotTemplate(subplots=[sbp1, sbp2, sbp3], small_subplots=False)
    >>> from valjean.javert import mpl
    >>> mplplt = mpl.MplPlot(pltnd)
    >>> fig, _ = mplplt.draw()


Strings as bins
---------------

It is possible to use strings as bins both in 1D and 2D plots. If strings are
too long on x-axis they will be represented vertically.

.. plot::
    :include-source:

    >>> from collections import OrderedDict
    >>> import numpy as np
    >>> from valjean.javert.templates import (PlotTemplate, CurveElements,
    ...                                       SubPlotElements)
    >>> bins = [np.array(['spam', 'egg', 'bacon']),
    ...         np.array(['beer', 'wine', 'milk', 'tea with milk and sugar'])]
    >>> axnames = ['x', 'y']
    >>> v2d = np.arange(12).reshape(3, 4)
    >>> v1d = np.arange(4)
    >>> lsplts = []
    >>> lsplts.append(SubPlotElements(
    ...     curves=[CurveElements(values=v2d, bins=bins, legend='Menus')],
    ...     axnames=['Meat', 'Drink', 'Associations'], ptype='2D'))
    >>> lsplts.append(SubPlotElements(
    ...     curves=[CurveElements(
    ...         values=v1d, bins=bins[1:], legend='', index=1)],
    ...     axnames=['Drink', 'Quantity'], ptype='1D'))
    >>> pltnd = PlotTemplate(subplots=lsplts)
    >>> from valjean.javert import mpl
    >>> mplplt = mpl.MplPlot(pltnd)
    >>> fig, _ = mplplt.draw()


Pie plot
--------

Pie plots can be done if requested. `ptype` should be equal to `'pie'`. Note
that the number of `axnames` still has to be N dim + 1, so 2. The first one is
the title of the plot, the second one the title of the legend. If the second
string is empty no title will be given to the legend.

.. plot::
    :include-source:

    >>> import numpy as np
    >>> from valjean.javert.templates import (PlotTemplate, CurveElements,
    ...                                       SubPlotElements)
    >>> ingredients = ['egg', 'spam', 'bacon', 'sausage', 'tomato', 'beans']
    >>> proportions = [0.1, 0.3, 0.25, 0.2, 0.05, 0.1]
    >>> curve = CurveElements(values=np.array(proportions),
    ...                       bins=[ingredients], legend='')
    >>> sbplt = SubPlotElements(curves=[curve],
    ...                         axnames=('Python pie', 'Ingredients'),
    ...                         ptype='pie')
    >>> pltpie = PlotTemplate(subplots=[sbplt])
    >>> from valjean.javert import mpl
    >>> mplplt = mpl.MplPlot(pltpie)
    >>> fig, _ = mplplt.draw()


Bar plots
---------

Bar plots can be used here with strings as labels in x-axis, like category
plots but no errors are expected here. Two kinds of bar plots are available:
side-by-side bars and stacked bars.

.. plot::
    :include-source:

    >>> import numpy as np
    >>> from valjean.javert.templates import (PlotTemplate, CurveElements,
    ...                                       SubPlotElements)
    >>> bins = [np.array(['spam', 'egg', 'bacon'])]
    >>> data = [np.array([1, 3, 4]),  np.array([2, 4, 5]),
    ...         np.array([5, 3, 2]), np.array([2, 3, 1])]
    >>> names = ['Terry', 'John' , 'Graham', 'Eric']
    >>> lcurves = []
    >>> for datum, name in zip(data, names):
    ...     lcurves.append(CurveElements(values=datum, bins=bins, legend=name))
    >>> speb = SubPlotElements(curves=lcurves,
    ...                        axnames=['ingredient', 'quantity'], ptype='bar')
    >>> spebs = SubPlotElements(curves=lcurves,
    ...                         axnames=['ingredient', 'quantity'],
    ...                         ptype='barstack')
    >>> pltbar = PlotTemplate(subplots=[speb, spebs], small_subplots=False)
    >>> from valjean.javert import mpl
    >>> mplplt = mpl.MplPlot(pltbar)
    >>> fig, _ = mplplt.draw()

Module API
----------
'''
import logging
from itertools import cycle, chain
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mplcol


LOGGER = logging.getLogger(__name__)


class MplStyle:
    '''Class to store style characteristics.'''

    def __init__(self, style=None, colors=None, mshape=None, mfill=None,
                 legends=None):
        '''Initialisation of the style.

        :param style: style from `matplotlib styles`_ or from user one, if
            ``None`` ``'default'`` is used
        :type style: str or None
        :param colors: colors from `matplotlib colors`_, if ``None`` CN are
            used
        :type colors: list(str) or None
        :param mshape: marker shapes from `matplotlib markers`_, if ``None`` a
            default sequence has been determined
        :type mshape: list(str) or None
        :param mfill: marker fill, ``None`` will use an alternance of
            ``'fill'`` and ``'none'``
        :type mfill: list(str) or None
        :param legends: keyword arguments from `legend documentation`_ to be
            passed to legend builder
        :type legends: dict or None

        An additional instance parameter is available and initialised in
        :class:`MplPlot` thanks to :meth:`styles_sequence`, ``fmts``. It builds
        the suite of styles of 1D curves from colors and markers.
        '''
        self.style = 'default' if style is None else style
        self.colors = (['C0', 'C1', 'C2', 'C3', 'C4', 'C5', 'C6', 'C7', 'C8',
                        'C9'] if colors is None else colors)
        self.markers_shape = (['s', 's', 'H', 'H', 'D', 'D', 'o', 'o', 'X',
                               'X', 'h', 'h', 'P', 'P', '8', '8']
                              if mshape is None else mshape)
        self.markers_fill = ['full', 'none'] if mfill is None else mfill
        self.legends = {} if legends is None else legends

    def styles_sequence(self, indices):
        '''Define the 1D style suite to be used for the 1D plots.

        :param list(int) indices: list of the curves style index
        :rtype: dict
        :returns: dictionary of curve styles indexed by curve index
        '''
        all_fmts = chain((('k', 'o', 'full'),),
                         zip(cycle(self.colors),
                             cycle(self.markers_shape),
                             cycle(self.markers_fill)))
        return dict(zip(indices, all_fmts))


class MplPlotException(Exception):
    '''Error raised if the plot cannot be made.'''


class MplPlot:
    '''Convert a :class:`~.templates.PlotTemplate` into a matplotlib plot.'''

    PTYPES = ('1D', '2D', 'pie', 'bar', 'barstack')

    def __init__(self, data, *, style=None):
        '''Construct a :class:`MplPlot` from the given
        :class:`~.templates.PlotTemplate`.

        Plots are initialized, drawn and finalized in :math:`draw`. Depending
        on the requested type of the subplots, ``ptype`` a 1D or a 2D plot will
        be done. Internal classes are called to draw each kind of available
        plots.

        Available types of plots are stored in the class variable ``PTYPES``.

        No plot for more than 2 dimensions are done.

        :param PlotTemplate data: the data to convert
        :param MplStyle style: the style to be used in the plot
        '''
        self.data = data
        self.style = MplStyle() if style is None else style
        plt.style.use(self.style.style)

    @staticmethod
    def figure_properties(data):
        '''Define figures properties like figsize or the grid specifications.

        :returns: dictionary of keyword arguments directly used by matplotlib
        :rtype: dict
        '''
        splts_kwargs = {}
        if 'nrows' not in data.backend_kw:
            splts_kwargs['nrows'] = data.nb_plots
        if 'ncols' in data.backend_kw:
            assert 'nrows' in data.backend_kw, (
                'nrows is also required when ncols is given')
        if 'figsize' not in data.backend_kw:
            splts_kwargs['figsize'] = (6.4, 6.4+2*(data.nb_plots-1))
            if data.small_subplots:
                splts_kwargs['figsize'] = (6.4, 4.8+1.2*(data.nb_plots-1))
            if 'pie' in [s.ptype for s in data.subplots]:
                splts_kwargs['figsize'] = (6.4, 4)
            if 'ncols' in data.backend_kw:
                splts_kwargs['figsize'] = (
                    6.4 + 6.4*(data.backend_kw['ncols']-1),
                    6.4 + 4.3*(data.backend_kw['nrows']-1))
        if data.backend_kw and data.small_subplots:
            LOGGER.warning('small_subplots=True may conflict with '
                           'backend-specific options passed via backend_kw')
        if 'gridspec_kw' not in data.backend_kw:
            if data.small_subplots:
                splts_kwargs['gridspec_kw'] = {
                    'height_ratios': [4] + [1]*(data.nb_plots-1)}
            if 'ncols' in data.backend_kw:
                splts_kwargs['gridspec_kw'] = {
                    'hspace': 0.4*data.backend_kw['nrows']}
        return splts_kwargs

    def initialize_figure(self):
        '''Construct the figure and its subplots.

        :rtype: tuple(matplotlib.figure.Figure, list(matplotlib.axes.Axes))
        '''
        fig, splts = plt.subplots(
            **self.figure_properties(self.data), **self.data.backend_kw,
            constrained_layout=True)
        if self.data.nb_plots == 1:
            splts = np.array([splts])
        return fig, splts

    def finalize_figure(self, splts):
        '''Finalize the figure.

        If ``suppress_xaxes`` is required in **data**, tick labels and label of
        the x-axis will be deleted on all subplots except the last one.

        If ``suppress_legends`` is required in **data**, legends will be
        deleted on all plots except the first one. If only one curve is
        represented in total legend is also deleted.


        :param list(matplotlib.axes.Axes) splts: the subplots
        '''
        for i, splt in enumerate(splts):
            if i != self.data.nb_plots-1 and self.data.suppress_xaxes:
                plt.setp(splt.get_xticklabels(), visible=False)
                splt.set_xlabel('')
            if ((i != 0 and self.data.suppress_legends)
                    or (sum(len(s.curves) for s in self.data.subplots) == 1
                        and self.data.subplots[i].ptype != 'pie')):
                plt.setp(splt.get_legend(), visible=False)

    def draw(self):
        '''Draw the plot.

        :rtype: tuple(matplotlib.figure.Figure, list(matplotlib.axes.Axes))
        '''
        if any(s.ptype not in MplPlot.PTYPES for s in self.data.subplots):
            ptypes = [s.ptype for s in self.data.subplots]
            raise MplPlotException(
                f"ptype from {ptypes} not taken into account. Expected ones "
                f"are {self.PTYPES}.")
        fig, splts = self.initialize_figure()
        fmts = self.style.styles_sequence(self.data.curves_index())
        for splt, sdat in zip(splts.flatten(), self.data.subplots):
            if sdat.ptype == '1D':
                mpl_plot = _MplPlot1D(sdat, self.style)
            elif sdat.ptype == '2D':
                mpl_plot = _MplPlot2D(sdat, self.style)
            elif sdat.ptype == 'pie':
                mpl_plot = _MplPie(sdat, self.style)
            elif sdat.ptype == 'bar':
                mpl_plot = _MplBar(sdat, self.style)
            elif sdat.ptype == 'barstack':
                mpl_plot = _MplBarStack(sdat, self.style)
            mpl_plot.draw(fig, splt, fmts=fmts)
        self.finalize_figure(splts)
        return fig, splts

    def save(self, name='fig.png'):
        '''Save the plot under the given name.

        :param str name: name of the output file. Expected extensions: png,
            pdf, svg, eps.
        '''
        LOGGER.debug('drawing figure %s', name)
        fig, _ = self.draw()
        if fig is not None:
            fig.savefig(name)
            plt.close(fig)


class _MplLegend:
    '''Class to store the legend content.'''

    def __init__(self, handle, label, index):
        '''Initialisation of :class:`_MplLegend`.

        :param handle: curve to be stored (if the curve needs to be drawn in
            twice a tuple should be given)
        :type handle: :obj:`matplotlib.pyplot.errorbar`
            or :obj:`tuple` (:obj:`matplotlib.pyplot.errorbar`)
        :param str label: the curve label
        :param int index: curve index, to identify its style
        '''
        self.handle = handle
        self.label = label
        self.index = index


class _MplPlot1D:
    '''Convert a :class:`~.templates.PlotTemplate` into a matplotlib plot.'''

    def __init__(self, data, style=None):
        '''Construct a :class:`_MplPlot1D` from the given
        :class:`~.templates.SubPlotElements`.

        :param SubPlotElements data: the data to convert.
        :param MplStyle style: style to be used in the subplot
        '''
        self.data = data
        self.mpl_style = MplStyle() if style is None else style
        self.legend = []

    def draw(self, _fig, splt, *_args, fmts, **_kwargs):
        '''Draw method.

        :param matplotlib.figure.Figure _fig: the current figure
        :param matplotlib.axes.Axes splt: the current subplot
        :param dict fmts: curves styles
        '''
        self.error_plots(splt, fmts)
        self.customize_plots(splt)

    def ierror_plot(self, splt, curve, data_fmt):
        '''Draw the plot with error bars on the plot (update the plot)

        :param matplotlib.axes.Axes splt: the current subplot
        :param CurveElements curve: data to plot
        :param tuple(str) data_fmt: format of the curve (color, marker shape,
            marker filling)
        '''
        LOGGER.debug("in ierror plot for plot %s on subplot %d",
                     curve.legend, 0)
        if curve.bins[0].size == curve.values.size+1:
            steps = splt.errorbar(
                curve.bins[0], np.append(curve.values, [np.nan]),
                linestyle='-', color=data_fmt[0], drawstyle='steps-post')
            markers = splt.errorbar(
                (curve.bins[0][1:] + curve.bins[0][:-1]) * 0.5,
                curve.values, yerr=curve.errors,
                color=data_fmt[0], marker=data_fmt[1], fillstyle=data_fmt[2],
                linestyle='')
            self.legend.append(_MplLegend((steps, markers), curve.legend,
                                          curve.index))
        else:
            eplt = splt.errorbar(
                curve.bins[0], curve.values, yerr=curve.errors,
                linestyle='--', color=data_fmt[0], marker=data_fmt[1],
                fillstyle=data_fmt[2])
            self.legend.append(_MplLegend(eplt, curve.legend, curve.index))

    def error_plots(self, splt, fmts):
        '''Plot errorbar plot (update the pyplot instance) and build the
        legend.

        :param matplotlib.axes.Axes splt: the current subplot
        :param dict fmts: curves styles
        '''
        for crv in self.data.curves:
            self.ierror_plot(splt, crv, fmts[crv.index])
        self._build_legend(splt)

    def _build_legend(self, splt):
        '''Build the legends and add them to the figures.

        An automatic number of columns is calculated, depending on the number
        of curves to be plotted on the subplot. It has been decided to add a
        new columns each 5 curves.

        :param matplotlib.axes.Axes splt: the current subplot
        '''
        ncol = len(self.data.curves) // 6 + 1
        splt.legend([lg.handle for lg in self.legend],
                    [lg.label for lg in self.legend],
                    ncol=ncol, **self.mpl_style.legends)

    def customize_plots(self, splt):
        '''Customize plots (scale, limits and lines).'''
        splt.set_xlabel(self.data.axnames[0])
        splt.set_ylabel(self.data.axnames[1])
        if self.data.curves[0].bins[0].dtype.kind == 'U':
            bmax = max(len(b) for b in self.data.curves[0].bins[0])
            if bmax * self.data.curves[0].bins[0].size > 60:
                splt.tick_params(axis='x', rotation=90)
        if self.data.attributes.limits is not None:
            splt.set_xlim(*self.data.attributes.limits[0])
        if self.data.attributes.logx:
            splt.set_xscale('log')
        if self.data.attributes.logy:
            splt.set_yscale('log')
        if self.data.attributes.lines:
            xlims, ylims = splt.get_xlim(), splt.get_ylim()
            for line in self.data.attributes.lines:
                if ('x' in line
                        and line['x'] > xlims[0] and line['x'] < xlims[1]):
                    splt.axvline(x=line['x'], c='grey', ls='--', lw=0.5)
                if ('y' in line
                        and line['y'] > ylims[0] and line['y'] < ylims[1]):
                    splt.axhline(y=line['y'], c='grey', ls='--', lw=0.5)


class _MplPlot2D:
    '''Convert a :class:`~.templates.SubPlotElements` into a 2D plot.'''

    def __init__(self, data, style=None):
        '''Construct a :class:`_MplPlot2D` from the given
        :class:`~.templates.SubPlotElements`.

        :param SubPlotElements data: the data to convert.
        :param MplStyle style: style to be used in the subplot
        '''
        LOGGER.debug('initialisation of MplPlot2D')
        self.data = data
        self.mpl_style = MplStyle() if style is None else style

    def draw(self, fig, splt, *_args, **_kwargs):
        '''Draw method.

        Remark: if the quantity represented is required in logarithmic scale,so
        the z-axis in logarithmic scale, it has to be done at the histogram
        declaration and not in the customization step.

        :param matplotlib.figure.Figure fig: the current figure
        :param matplotlib.axes.Axes splt: the current subplot
        '''
        self.twod_plots(fig, splt)
        self.customize_plots(splt)

    @staticmethod
    def broadcast_bin_centers(curve):
        '''Calculate bin centers if edges are given and broadcast all bins:
        build the (x, y) grid for all bins.

        :param CurveElements curve: data to plot
        :rtype: list(numpy.ndarray), list(numpy.ndarray), list(numpy.ndarray)
        :returns: grid, bin edges, bin centers
        '''
        bins, cbins, rsbins = [], [], []
        for idim, tbin in enumerate(curve.bins):
            shape = ([curve.values.shape[idim]]
                     + [1] * (curve.values.ndim - 1 - idim))
            if tbin.dtype.kind == 'U':
                cbins.append(np.arange(tbin.size)+0.5)
                bins.append(np.arange(tbin.size+1))
            elif tbin.size == curve.values.shape[idim]:
                # here tbin represents the bin centers
                assert tbin.size > 1, 'The axis should not be trivial, > 1 bin'
                cbins.append(tbin)
                hbwidth = np.ediff1d(tbin) * 0.5
                rbins = np.concatenate([[tbin[0]-hbwidth[0]],
                                        tbin[:-1]+hbwidth,
                                        [tbin[-1]+hbwidth[-1]]])
                bins.append(rbins)
            else:
                # here tbin represents the bin edges
                cbins.append((tbin[1:] + tbin[:-1]) * 0.5)
                bins.append(tbin)
            rsbins.append(np.array(cbins[-1]).reshape(shape))
        gbins = np.broadcast_arrays(*rsbins)
        return gbins, bins, cbins

    def itwod_plot(self, fig, splt, curve, axnames, norm):
        '''Draw the 2D distribution on the ith subplot.

        :param matplotlib.figure.Figure fig: the current figure
        :param matplotlib.axes.Axes splt: the current subplot
        :param CurveElements curve: data to plot
        :param int iplot: number of the subplot
        :param norm: function corresponding to the chosen
            normalisation and scale (linear or logarithmic)
        :type norm: function from :obj:`matplotlib.colors`
        '''
        gbins, bins, cbins = self.broadcast_bin_centers(curve)
        h2d = splt.hist2d(
            gbins[0].flatten(), gbins[1].flatten(),
            bins=bins, norm=norm, weights=curve.values.flatten())
        cbar = fig.colorbar(h2d[3], ax=splt)
        splt.set_xlabel(axnames[0])
        if curve.bins[0].size == curve.values.shape[0]:
            splt.set_xticks(cbins[0])
            splt.set_xticklabels(list(curve.bins[0]))
            if curve.bins[0].dtype.kind == 'U':
                bmax = max(len(b) for b in self.data.curves[0].bins[0])
                if bmax * self.data.curves[0].bins[0].size > 60:
                    splt.tick_params(axis='x', rotation=90)
        splt.set_ylabel(axnames[1])
        if curve.bins[1].size == curve.values.shape[1]:
            splt.set_yticks(cbins[1])
            splt.set_yticklabels(list(curve.bins[1]))
        cbar.set_label(axnames[2])
        if curve.legend:
            splt.set_title(curve.legend)

    def twod_plots(self, fig, splt):
        '''Build 2D plots.

        :param matplotlib.figure.Figure fig: the current figure
        :param matplotlib.axes.Axes splt: the current subplot
        '''
        for crv in self.data.curves:
            self.itwod_plot(
                fig, splt, crv, self.data.axnames,
                norm=(mplcol.LogNorm() if self.data.attributes.logz
                      else mplcol.Normalize()))

    def customize_plots(self, splt):
        '''Customize plots (scale and limit).

        :param matplotlib.figure.Figure fig: the current figure
        :param matplotlib.axes.Axes splt: the current subplot
        '''
        if self.data.attributes.limits is not None:
            splt.set_xlim(*self.data.attributes.limits[0])
            splt.set_ylim(*self.data.attributes.limits[1])
        if self.data.attributes.logx:
            splt.set_xscale('log')
        if self.data.attributes.logy:
            splt.set_yscale('log')
        if self.data.curves[0].bins[0].dtype.kind == 'U':
            bmax = max(len(b) for b in self.data.curves[0].bins[0])
            if bmax * self.data.curves[0].bins[0].size > 60:
                splt.tick_params(axis='x', rotation=90)


class _MplPie:
    '''Convert a :class:`~.templates.PlotTemplate` into a matplotlib pie chart.
    '''

    def __init__(self, data, _style=None):
        '''Construct a :class:`_MplPie` from the given
        :class:`~.templates.SubPlotElements`.

        :param SubPlotElements data: the data to convert.
        :param MplStyle style: style to be used in the subplot
        '''
        self.data = data
        self.fig, self.splt = None, None

    def draw(self, _fig, splt, *_args, **_kwargs):
        '''Draw method.

        :param matplotlib.figure.Figure _fig: the current figure
        :param matplotlib.axes.Axes splt: the current subplot
        '''
        self.pie_chart(splt)

    def pie_chart(self, splt):
        '''Prepare the pie chart.

        :param matplotlib.axes.Axes splt: the current subplot
        '''
        curve = self.data.curves[0]
        wedges, _, _ = splt.pie(curve.values, autopct='%1.1f%%')
        splt.set_title(self.data.axnames[0])
        splt.legend(wedges, curve.bins[0],
                    loc='center right', bbox_to_anchor=(1.3, 0.5),
                    title=self.data.axnames[1])


class _MplBar:
    '''Convert a :class:`~.templates.PlotTemplate` into a matplotlib bar plot.
    '''

    def __init__(self, data, _style=None):
        '''Construct a :class:`_MplBar` from the given
        :class:`~.templates.SubPlotElements`.

        :param SubPlotElements data: the data to convert.
        :param MplStyle style: style to be used in the subplot
        '''
        self.data = data

    def draw(self, _fig, splt, *_args, **_kwargs):
        '''Draw method.

        :param matplotlib.figure.Figure _fig: the current figure
        :param matplotlib.axes.Axes splt: the current subplot
        '''
        self.bar_plot(splt)

    def bar_plot(self, splt):
        '''Prepare the bar plot.'''
        nb_curves = len(self.data.curves)
        width = 0.8 / nb_curves
        x = np.arange(self.data.curves[0].values.size)
        bot = np.zeros_like(self.data.curves[0].values)
        for icrv, curve in enumerate(self.data.curves):
            splt.bar(x + (icrv-(nb_curves-1)/2)*width, curve.values, width,
                     label=curve.legend)
            bot += curve.values
        splt.set_xlabel(self.data.axnames[0])
        splt.set_ylabel(self.data.axnames[1])
        splt.set_xticks(x)
        splt.set_xticklabels(self.data.curves[0].bins[0])
        splt.legend()


class _MplBarStack:
    '''Convert a :class:`~.templates.PlotTemplate` into a matplotlib bar plot.
    '''

    def __init__(self, data, _style=None):
        '''Construct a :class:`_MplBarStack` from the given
        :class:`~.templates.SubPlotElements`.

        :param SubPlotElements data: the data to convert.
        :param MplStyle style: style to be used in the subplot
        '''
        self.data = data

    def draw(self, _fig, splt, *_args, **_kwargs):
        '''Draw method.

        :param matplotlib.figure.Figure _fig: the current figure
        :param matplotlib.axes.Axes splt: the current subplot
        '''
        self.bar_plot(splt)

    def bar_plot(self, splt):
        '''Prepare the bar plot.'''
        nb_curves = len(self.data.curves)
        width = 0.8 / nb_curves
        bot = np.zeros_like(self.data.curves[0].values)
        for curve in self.data.curves:
            splt.bar(self.data.curves[0].bins[0], curve.values,
                     width*nb_curves, label=curve.legend, bottom=bot)
            bot += curve.values
        splt.set_xlabel(self.data.axnames[0])
        splt.set_ylabel(self.data.axnames[1])
        splt.legend()



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''The purpose of this module is to define the :class:`EvalTestTask` class, a
task that evaluates a collection of :class:`~.gavroche.test.Test` objects and
transforms them into :class:`~.gavroche.test.TestResult` objects, which can be
subsequently processed for inclusion in a test report.
'''

import logging
from pathlib import Path

from ..path import ensure, sanitize_filename
from ..cosette.task import TaskStatus
from ..cosette.use import from_env
from ..cosette.pythontask import PythonTask
from ..gavroche.test import Test, TestResultFailed


LOGGER = logging.getLogger(__name__)


def actually_eval_test(test):
    '''Actually perform test evaluation.

    :returns: the result of test evaluation.
    :rtype: TestResult
    :raises TestResultFailed: if the underlying test raises any exception.
    '''
    try:
        res = test.evaluate()
    except Exception as ex:  # pylint: disable=broad-except
        LOGGER.error('The test failed with error: \n%s', ex)
        return TestResultFailed(test, ex)
    return res


class EvalTestTask(PythonTask):
    '''Class that evaluates a list of tests and stores the resulting
    :class:`~.TestResult` objects in the environment.'''

    @classmethod
    def from_test_task(cls, test_task, name=None):
        '''This method instantiates an :class:`EvalTestTask` that will evaluate
        all the tests generated by a given task.

        :param Task test_task: a task that is expected to generate a list of
                               tests as a result.
        '''
        test_task_name = test_task.name
        if name is None:
            eval_task_name = test_task_name + '.eval'
        else:
            eval_task_name = name
        return cls(eval_task_name, test_task_name, deps=[test_task])

    def __init__(self, name, test_task_name, *, deps=None, soft_deps=None):
        '''Direct instantiation of an :class:`EvalTestTask`.

        :param str name: the name of this task.
        :param str test_task_name: the name of the task that generated the
                                   tests.
        :param deps: the list of dependencies for this task.
        :type deps: list(Task) or None
        :param soft_deps: the list of soft dependencies for this task.
        :type soft_deps: list(Task) or None
        '''
        def evaluate(*, env, config):
            tests = from_env(env=env, task_name=test_task_name, key='result')
            if not isinstance(tests, list):
                raise TypeError('Expected a list of tests in EvalTestTask '
                                f'{self.name!r}; got a {type(tests)} instead')
            for test in tests:
                if not isinstance(test, Test):
                    raise TypeError('Expected a list of tests in EvalTestTask '
                                    f'{self.name!r}, but one of the list '
                                    'elements is a {type(test)}')

            results = [actually_eval_test(test) for test in tests]

            output_dir = Path(config.query('path', 'output-root'),
                              sanitize_filename(self.name))
            ensure(output_dir, is_dir=True)
            env_up = {self.name: {'result': results,
                                  'output_dir': str(output_dir)}}
            status = TaskStatus.DONE
            return env_up, status

        super().__init__(name, evaluate, deps=deps, soft_deps=soft_deps,
                         env_kwarg='env', config_kwarg='config')


def evaluate_tests(test_fn, name=None):
    '''Create an :class:`EvalTestTask` objects for the given test.

    :param test_fn: a function that produces tests, wrapped in a
        :class:`~valjean.cosette.use.Use` decorator.
    :type test_fn: valjean.cosette.use.Use
    :param str name: the name of the :class:`EvalTestTask` object.
    :returns: the task that evaluates your tests.
    :rtype: EvalTestTask
    '''
    return EvalTestTask.from_test_task(test_fn.get_task(), name=name)
